from __future__ import annotations

import pytest

from cognite_toolkit._cdf_tk._parameters import (
    ANY_STR,
    ANYTHING,
    ParameterSet,
    ParameterSpec,
    ParameterSpecSet,
    ParameterValue,
)

SPACE_SPEC = ParameterSpecSet(
    {
        ParameterSpec(("space",), frozenset({"str"}), True, False),
        ParameterSpec(("description",), frozenset({"str"}), False, True),
        ParameterSpec(("name",), frozenset({"str"}), False, True),
    }
)

DATASET_SPEC = ParameterSpecSet(
    {
        ParameterSpec(path=("description",), types=frozenset({"str"}), is_required=False, _is_nullable=True),
        ParameterSpec(path=("name",), types=frozenset({"str"}), is_required=False, _is_nullable=True),
        ParameterSpec(path=("externalId",), types=frozenset({"str"}), is_required=False, _is_nullable=True),
        ParameterSpec(path=("metadata", ANY_STR), types=frozenset({"str"}), is_required=False, _is_nullable=False),
        ParameterSpec(path=("metadata",), types=frozenset({"dict"}), is_required=False, _is_nullable=True),
        ParameterSpec(path=("writeProtected",), types=frozenset({"bool"}), is_required=False, _is_nullable=True),
    }
)


class TestSetOperations:
    @pytest.mark.parametrize(
        "first, second, expected",
        [
            pytest.param(
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("description",), "str", "description"),
                        ParameterValue(("name",), "str", "name"),
                    }
                ),
                SPACE_SPEC,
                ParameterSet[ParameterValue]({}),
                id="Value - Spec no difference",
            ),
            pytest.param(
                SPACE_SPEC,
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("description",), "str", "description"),
                        ParameterValue(("name",), "str", "name"),
                    }
                ),
                ParameterSpecSet({}),
                id="Spec - Value no difference",
            ),
            pytest.param(
                SPACE_SPEC.required(),
                ParameterSet[ParameterValue](),
                ParameterSpecSet({ParameterSpec(("space",), frozenset({"str"}), True, False)}),
                id="Missing required",
            ),
            pytest.param(
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("wrong_name",), "str", "description"),
                    }
                ),
                SPACE_SPEC,
                ParameterSet[ParameterValue]({ParameterValue(("wrong_name",), "str", "description")}),
                id="One unexpected value",
            ),
            pytest.param(
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(path=("name",), type="str", value="B_5rVE|'2@vQnR!)%s|Cnoz^(.yY%`[3HfoXFO6&\\"),
                        ParameterValue(path=("writeProtected",), type="bool", value=True),
                        ParameterValue(path=("description",), type="str", value="e(Ox6G"),
                        ParameterValue(
                            path=(
                                "metadata",
                                '"u&SjvpP%<aS*wdOi$4Cuy2_p,?V2f9xN:*W/I$6<ZntaN=l9yKWE-To^sJ=`ltyykBN|]zv^(B-P)y%.h.")',
                            ),
                            type="str",
                            value="#7x,",
                        ),
                        ParameterValue(path=("metadata",), type="dict", value=None),
                        ParameterValue(
                            path=("externalId",), type="str", value="WTNVF3N27L9HX6D3OCZMO22X3VQ82LY6QJHCSWUBOXWYWHKQ62"
                        ),
                    }
                ),
                DATASET_SPEC,
                ParameterSet[ParameterValue](),
                id="No difference DataSet, with AnyStr",
            ),
            pytest.param(
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(path=("name",), type="str", value="name"),
                        ParameterValue(
                            path=(
                                "metadata",
                                "some key",
                            ),
                            type="str",
                            value="#7x,",
                        ),
                        ParameterValue(
                            path=("metadata", "some key", "type"),
                            type="str",
                            value="json",
                        ),
                        ParameterValue(path=("metadata",), type="dict", value=None),
                    }
                ),
                ParameterSpecSet(
                    {
                        ParameterSpec(
                            path=("metadata", ANYTHING),
                            types=frozenset({"dict"}),
                            is_required=False,
                            _is_nullable=False,
                        ),
                    }
                ),
                ParameterSet[ParameterValue](
                    {
                        ParameterValue(path=("name",), type="str", value="name"),
                        ParameterValue(path=("metadata",), type="dict", value=None),
                    }
                ),
                id="Diff with Anything in spec",
            ),
        ],
    )
    def test_difference(self, first: ParameterSet, second: ParameterSet, expected: ParameterSet) -> None:
        actual = first - second

        assert sorted(actual) == sorted(expected)

    @pytest.mark.parametrize(
        "my_set, path, expected",
        [
            pytest.param(
                SPACE_SPEC,
                ("space",),
                ParameterSpecSet({ParameterSpec(("space",), frozenset({"str"}), True, False)}),
                id="Subset tuple",
            ),
            pytest.param(
                ParameterSpecSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("properties", "type"), "str", "description"),
                        ParameterValue(("name",), "str", "name"),
                    }
                ),
                1,
                ParameterSpecSet[ParameterValue](
                    {
                        ParameterValue(("space",), "str", "space"),
                        ParameterValue(("name",), "str", "name"),
                    }
                ),
                id="One unexpected value",
            ),
        ],
    )
    def test_subset(self, my_set: ParameterSet, path: int | tuple[str | int, ...], expected: ParameterSet) -> None:
        actual = my_set.subset(path)

        assert sorted(actual) == sorted(expected)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    RobotCapability,
    RobotCapabilityList,
    RobotCapabilityWrite,
    _RobotCapabilityUpdate,
)

from .utlis import tmp_disable_gzip


class CapabilitiesAPI(APIClient):
    _RESOURCE_PATH = "/robotics/capabilities"

    @overload
    def __call__(self) -> Iterator[RobotCapability]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[RobotCapabilityList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[RobotCapability] | Iterator[RobotCapabilityList]:
        """Iterate over robot capabilities.

        Args:
            chunk_size: The number of robot capabilities to return in each chunk. None will return all robot capabilities.

        Yields:
            RobotCapability or RobotCapabilityList

        """
        return self._list_generator(
            method="GET", resource_cls=RobotCapability, list_cls=RobotCapabilityList, chunk_size=chunk_size
        )

    def __iter__(self) -> Iterator[RobotCapability]:
        return self.__call__()

    @overload
    def create(self, capability: RobotCapabilityWrite) -> RobotCapability: ...

    @overload
    def create(self, capability: Sequence[RobotCapabilityWrite]) -> RobotCapabilityList: ...

    def create(
        self, capability: RobotCapabilityWrite | Sequence[RobotCapabilityWrite]
    ) -> RobotCapability | RobotCapabilityList:
        """Create a new robot capability.

        Args:
            capability: RobotCapabilityWrite or list of RobotCapabilityWrite.

        Returns:
            RobotCapability object.

        """
        with tmp_disable_gzip():
            return self._create_multiple(
                list_cls=RobotCapabilityList,
                resource_cls=RobotCapability,
                items=capability,
                input_resource_cls=RobotCapabilityWrite,
            )

    @overload
    def retrieve(self, external_id: str) -> RobotCapability | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> RobotCapabilityList: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> RobotCapability | None | RobotCapabilityList:
        """Retrieve a robot capability.

        Args:
            external_id: External id of the robot capability.

        Returns:
            RobotCapability object.

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            return self._retrieve_multiple(
                identifiers=identifiers,
                resource_cls=RobotCapability,
                list_cls=RobotCapabilityList,
            )

    @overload
    def update(self, capability: RobotCapabilityWrite) -> RobotCapability: ...

    @overload
    def update(self, capability: Sequence[RobotCapabilityWrite]) -> RobotCapabilityList: ...

    def update(
        self, capability: RobotCapabilityWrite | Sequence[RobotCapabilityWrite]
    ) -> RobotCapability | RobotCapabilityList:
        """Update a robot capability.

        Args:
            capability: RobotCapabilityWrite or list of RobotCapabilityWrite.

        Returns:
            RobotCapability object.

        """
        with tmp_disable_gzip():
            return self._update_multiple(
                items=capability,
                resource_cls=RobotCapability,
                list_cls=RobotCapabilityList,
                update_cls=_RobotCapabilityUpdate,
            )

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete a robot capability.

        Args:
            external_id: External id of the robot capability.

        Returns:
            None

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            self._delete_multiple(identifiers=identifiers, wrap_ids=True)

    def list(self) -> RobotCapabilityList:
        """List robot capabilities.

        Returns:
            RobotCapabilityList

        """
        with tmp_disable_gzip():
            return self._list(method="GET", resource_cls=RobotCapability, list_cls=RobotCapabilityList)



from functools import total_ordering

BASE_TYPES = {t.__name__ for t in (str, int, float, bool)}
CONTAINER_TYPES = {t.__name__ for t in (list, dict)}
TYPES = BASE_TYPES | CONTAINER_TYPES


@total_ordering
class _AnyInt(str):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            return True
        elif isinstance(other, str):
            return False
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, int):
            return True
        elif isinstance(other, str):
            return True
        return NotImplemented

    def __hash__(self) -> int:
        return id(ANY_INT)

    def __str__(self) -> str:
        return "AnyInt"

    def __repr__(self) -> str:
        return "AnyInt"


@total_ordering
class _AnyStr(str):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return True
        elif isinstance(other, int):
            return False
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str):
            return True
        elif isinstance(other, int):
            return False
        return NotImplemented

    def __hash__(self) -> int:
        return id(ANY_STR)

    def __str__(self) -> str:
        return "AnyStr"

    def __repr__(self) -> str:
        return "AnyStr"


@total_ordering
class _Anything(str):
    def __eq__(self, other: object) -> bool:
        return True

    def __lt__(self, other: object) -> bool:
        return True

    def __hash__(self) -> int:
        return id(ANYTHING)

    def __str__(self) -> str:
        return "Anything"

    def __repr__(self) -> str:
        return "Anything"


ANY_INT = _AnyInt()
ANY_STR = _AnyStr()
ANYTHING = _Anything()

SINGLETONS = {ANY_STR, ANY_INT, ANYTHING}



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any
from unittest.mock import MagicMock

from cognite.client.testing import CogniteClientMock

from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient

from .api.dml import DMLAPI
from .api.location_filters import LocationFiltersAPI
from .api.lookup import (
    AssetLookUpAPI,
    DataSetLookUpAPI,
    ExtractionPipelineLookUpAPI,
    LocationFiltersLookUpAPI,
    LookUpGroup,
    SecurityCategoriesLookUpAPI,
    TimeSeriesLookUpAPI,
)
from .api.robotics import RoboticsAPI
from .api.robotics.capabilities import CapabilitiesAPI
from .api.robotics.data_postprocessing import DataPostProcessingAPI
from .api.robotics.frames import FramesAPI
from .api.robotics.locations import LocationsAPI as RoboticsLocationsAPI
from .api.robotics.maps import MapsAPI
from .api.statistics import StatisticsAPI
from .api.verify import VerifyAPI


class ToolkitClientMock(CogniteClientMock):
    """Mock for ToolkitClient object

    All APIs are replaced with specked MagicMock objects.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "parent" in kwargs:
            super().__init__(*args, **kwargs)
            return None
        super().__init__(*args, **kwargs)
        # Developer note:
        # - Please add your mocked APIs in chronological order
        # - For nested APIs:
        #   - Add spacing above and below
        #   - Use `spec=MyAPI` only for "top level"
        #   - Use `spec_set=MyNestedAPI` for all nested APIs

        self.dml = MagicMock(spec_set=DMLAPI)
        self.location_filters = MagicMock(spec_set=LocationFiltersAPI)
        self.lookup = MagicMock(spec=LookUpGroup)
        self.lookup.data_sets = MagicMock(spec_set=DataSetLookUpAPI)
        self.lookup.assets = MagicMock(spec_set=AssetLookUpAPI)
        self.lookup.time_series = MagicMock(spec_set=TimeSeriesLookUpAPI)
        self.lookup.security_categories = MagicMock(spec_set=SecurityCategoriesLookUpAPI)
        self.lookup.location_filters = MagicMock(spec_set=LocationFiltersLookUpAPI)
        self.lookup.extraction_pipelines = MagicMock(spec_set=ExtractionPipelineLookUpAPI)

        self.robotics = MagicMock()
        self.robotics.robots = MagicMock(spec=RoboticsAPI)
        self.robotics.data_postprocessing = MagicMock(spec_set=DataPostProcessingAPI)
        self.robotics.locations = MagicMock(spec_set=RoboticsLocationsAPI)
        self.robotics.frames = MagicMock(spec_set=FramesAPI)
        self.robotics.maps = MagicMock(spec_set=MapsAPI)
        self.robotics.capabilities = MagicMock(spec_set=CapabilitiesAPI)

        self.data_modeling.statistics = MagicMock(spec_set=StatisticsAPI)

        self.verify = MagicMock(spec_set=VerifyAPI)


@contextmanager
def monkeypatch_toolkit_client() -> Iterator[ToolkitClientMock]:
    toolkit_client_mock = ToolkitClientMock()
    ToolkitClient.__new__ = lambda *args, **kwargs: toolkit_client_mock  # type: ignore[method-assign]
    yield toolkit_client_mock
    ToolkitClient.__new__ = lambda cls, *args, **kwargs: object.__new__(cls)  # type: ignore[method-assign]



import json
import socket

import pytest
import responses
from cognite.client import global_config
from cognite.client.credentials import Token
from cognite.client.data_classes.data_modeling import NodeApply, NodeOrEdgeData, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteTimeSeriesApply
from cognite.client.exceptions import CogniteAPIError, CogniteConnectionError, CogniteReadTimeout

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig

BASE_URL = "http://blabla.cognitedata.com"
TOKEN_URL = "https://test.com/token"


@pytest.fixture
def toolkit_config():
    return ToolkitClientConfig(
        client_name="test-client",
        project="test-project",
        base_url=BASE_URL,
        max_workers=1,
        timeout=10,
        credentials=Token("abc"),
    )


@pytest.fixture
def disable_gzip():
    old = global_config.disable_gzip
    global_config.disable_gzip = True
    yield
    global_config.disable_gzip = old


@pytest.fixture
def max_retries_2():
    old = global_config.max_retries
    global_config.max_retries = 2
    yield
    global_config.max_retries = old


@pytest.fixture()
def some_timeseries() -> CogniteTimeSeriesApply:
    return CogniteTimeSeriesApply(
        space="some_space",
        external_id="test_time_series",
        is_step=False,
        time_series_type="numeric",
    )


class TestInstances:
    @pytest.mark.usefixtures("disable_gzip")
    def test_apply_fast_429_status_split(self, toolkit_config: ToolkitClientConfig) -> None:
        instances = [
            CogniteTimeSeriesApply(
                space="some_space",
                external_id=f"test_time_series_{i}",
                is_step=False,
                time_series_type="numeric",
            )
            for i in range(2)
        ]
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=429,
                json={"error": "Too many items"},
            )
            for ts in instances:
                rsps.add(
                    responses.POST,
                    url,
                    status=200,
                    json={
                        "items": [
                            {
                                "instanceType": "node",
                                "space": ts.space,
                                "externalId": ts.external_id,
                                "version": 1,
                                "wasModified": False,
                                "createdTime": 0,
                                "lastUpdatedTime": 0,
                            }
                        ]
                    },
                )
            client = ToolkitClient(config=toolkit_config)

            results = client.data_modeling.instances.apply_fast(instances)

            assert len(results) == 2
            assert len(rsps.calls) >= 3
            failed_call = rsps.calls[-3]
            assert failed_call.response.status_code == 429
            assert len(json.loads(failed_call.request.body)["items"]) == 2
            for call in rsps.calls[-2:]:
                assert call.response.status_code == 200
                assert len(json.loads(call.request.body)["items"]) == 1

    @pytest.mark.usefixtures("disable_gzip")
    def test_apply_fast_429_single_instance(
        self, toolkit_config: ToolkitClientConfig, some_timeseries: CogniteTimeSeriesApply
    ) -> None:
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=429,
                json={"error": "Too many items"},
            )
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={
                    "items": [
                        {
                            "instanceType": "node",
                            "space": some_timeseries.space,
                            "externalId": some_timeseries.external_id,
                            "version": 1,
                            "wasModified": False,
                            "createdTime": 0,
                            "lastUpdatedTime": 0,
                        }
                    ]
                },
            )
            client = ToolkitClient(config=toolkit_config)

            results = client.data_modeling.instances.apply_fast([some_timeseries])

            assert len(results) == 1
            assert len(rsps.calls) >= 2
            failed_call = rsps.calls[-2]
            assert failed_call.response.status_code == 429
            assert len(json.loads(failed_call.request.body)["items"]) == 1
            last_call = rsps.calls[-1]
            assert last_call.response.status_code == 200
            assert len(json.loads(last_call.request.body)["items"]) == 1

    @pytest.mark.usefixtures("max_retries_2")
    @pytest.mark.parametrize(
        "args, expected_exception",
        [
            pytest.param(
                dict(status=429, json={"error": "Server is busy, please try again later"}),
                CogniteAPIError,
                id="429 Too Many Requests",
            ),
            pytest.param(
                dict(body=ConnectionRefusedError("Connection refused")), CogniteConnectionError, id="Connection refused"
            ),
            pytest.param(dict(body=socket.timeout("timed out")), CogniteReadTimeout, id="Connection timed out"),
        ],
    )
    def test_apply_fast_raise(
        self,
        args: dict[str, object],
        expected_exception: type[Exception],
        toolkit_config: ToolkitClientConfig,
        some_timeseries: CogniteTimeSeriesApply,
    ) -> None:
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/models/instances"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                **args,
            )
            client = ToolkitClient(config=toolkit_config)
            with pytest.raises(expected_exception) as _:
                _ = client.data_modeling.instances.apply_fast([some_timeseries])

    def test_apply_fast_invalid_json(self, toolkit_config: ToolkitClientConfig) -> None:
        node = NodeApply(
            space="some_space",
            external_id="test_time_series",
            sources=[
                NodeOrEdgeData(source=ViewId("some_space", "some_view", "v1"), properties={"my_number": float("nan")})
            ],
        )
        client = ToolkitClient(config=toolkit_config)

        with pytest.raises(ValueError) as exc_info:
            _ = client.data_modeling.instances.apply_fast([node])

        assert "Out of range float values are not JSON compliant." in str(exc_info.value)

    def test_apply_fast_empty(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config)
        result = client.data_modeling.instances.apply_fast([])
        assert len(result) == 0



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    Map,
    MapList,
    MapWrite,
    _MapUpdate,
)

from .utlis import tmp_disable_gzip


class MapsAPI(APIClient):
    _RESOURCE_PATH = "/robotics/maps"

    @overload
    def __call__(self) -> Iterator[Map]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[MapList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[Map] | Iterator[MapList]:
        """Iterate over robot maps.

        Args:
            chunk_size: The number of robot maps to return in each chunk. None will return all robot maps.

        Yields:
            Map or MapList

        """
        return self._list_generator(method="GET", resource_cls=Map, list_cls=MapList, chunk_size=chunk_size)

    def __iter__(self) -> Iterator[Map]:
        return self.__call__()

    @overload
    def create(self, map: MapWrite) -> Map: ...

    @overload
    def create(self, map: Sequence[MapWrite]) -> MapList: ...

    def create(self, map: MapWrite | Sequence[MapWrite]) -> Map | MapList:
        """Create a new robot map.

        Args:
            map: MapWrite or list of MapWrite.

        Returns:
            Map object.

        """
        with tmp_disable_gzip():
            return self._create_multiple(
                list_cls=MapList,
                resource_cls=Map,
                items=map,
                input_resource_cls=MapWrite,
            )

    @overload
    def retrieve(self, external_id: str) -> Map | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str]) -> MapList: ...

    def retrieve(self, external_id: str | SequenceNotStr[str]) -> Map | None | MapList:
        """Retrieve a robot map.

        Args:
            external_id: External id of the robot map.

        Returns:
            Map object.

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            return self._retrieve_multiple(
                identifiers=identifiers,
                resource_cls=Map,
                list_cls=MapList,
            )

    @overload
    def update(self, map: MapWrite) -> Map: ...

    @overload
    def update(self, map: Sequence[MapWrite]) -> MapList: ...

    def update(self, map: MapWrite | Sequence[MapWrite]) -> Map | MapList:
        """Update a robot map.

        Args:
            map: MapWrite or list of MapWrite.

        Returns:
            Map object.

        """
        with tmp_disable_gzip():
            return self._update_multiple(
                items=map,
                resource_cls=Map,
                list_cls=MapList,
                update_cls=_MapUpdate,
            )

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete a robot map.

        Args:
            external_id: External id of the robot map.

        Returns:
            None

        """
        identifiers = IdentifierSequence.load(external_ids=external_id)
        with tmp_disable_gzip():
            self._delete_multiple(identifiers=identifiers, wrap_ids=True)

    def list(self) -> MapList:
        """List robot maps.

        Returns:
            MapList

        """
        with tmp_disable_gzip():
            return self._list(method="GET", resource_cls=Map, list_cls=MapList)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from common.dataclass.common import Dataclass


@dataclass
class File(Dataclass):
    fileId: int
    fileExternalId: Optional[str] = None


@dataclass
class RequestCustomModelPrediction(Dataclass):
    items: list[File]
    modelFile: File
    threshold: float


@dataclass
class Vertex(Dataclass):
    x: float
    y: float


@dataclass
class Region(Dataclass):
    shape: str  # points, rectangle, polyline, polygon
    vertices: list[Vertex]


@dataclass
class VisionAnnotation(Dataclass):
    text: str
    confidence: Optional[float] = None
    region: Optional[Region] = None


@dataclass
class Item(Dataclass):
    fileId: int
    annotations: Optional[list[VisionAnnotation]] = None
    fileExternalId: Optional[str] = None
    width: Optional[float] = None
    height: Optional[float] = None


@dataclass
class FailedBatchSchema(Dataclass):
    errorMessage: Optional[str] = None
    items: Optional[list[File]] = None


@dataclass
class ResponseCustomModelPrediction(Dataclass):
    status: str  # "Queued" "Running" "Completed" "Failed"
    createdTime: int
    startTime: int
    statusTime: int
    jobId: int
    items: list[Item]
    modelFile: File
    threshold: Optional[float] = None
    failedItems: Optional[list[FailedBatchSchema]] = None



import copy
from collections.abc import Callable, Iterable
from typing import Any

from cognite_toolkit._cdf_tk.builders import Builder
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    ModuleLocation,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitYAMLFormatError
from cognite_toolkit._cdf_tk.loaders import CogniteFileLoader, FileLoader, FileMetadataLoader
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning, ToolkitWarning


class FileBuilder(Builder):
    _resource_folder = FileMetadataLoader.folder_name
    template_pattern = "$FILENAME"

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | list[ToolkitWarning]]:
        for source_file in source_files:
            loaded = source_file.loaded
            if loaded is None:
                continue

            loader, warning = self._get_loader(source_file.source.path)
            if loader is None:
                if warning is not None:
                    yield [warning]
                continue
            if loader in {FileMetadataLoader, CogniteFileLoader}:
                loaded = self._expand_file_metadata(loaded, module, console)
            destination_path = self._create_destination_path(source_file.source.path, loader.kind)

            yield BuildDestinationFile(
                path=destination_path,
                loaded=loaded,
                loader=loader,
                source=source_file.source,
                extra_sources=None,
            )

    @classmethod
    def _expand_file_metadata(
        cls,
        raw_list: list[dict[str, Any]] | dict[str, Any],
        module: ModuleLocation,
        console: Callable[[str], None] | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        is_file_template = (
            isinstance(raw_list, list)
            and len(raw_list) == 1
            and cls.template_pattern in raw_list[0].get("externalId", "")
        )
        if not is_file_template:
            if (isinstance(raw_list, dict) and cls.template_pattern in raw_list.get("externalId", "")) or (
                isinstance(raw_list, list)
                and any(cls.template_pattern in entry.get("externalId", "") for entry in raw_list)
            ):
                raw_type = "dictionary" if isinstance(raw_list, dict) else "list with multiple entries"
                LowSeverityWarning(
                    f"Invalid file template {cls.template_pattern!r} usage detected in {module.relative_path.as_posix()!r}.\n"
                    f"The file template is expected in a list with a single entry, but got {raw_type}."
                ).print_warning()

            return raw_list
        if not (isinstance(raw_list, list) and raw_list and isinstance(raw_list[0], dict)):
            raise ToolkitYAMLFormatError(
                f"Expected a list with a single dictionary in the file metadata file {module.dir}, "
                f"but got {type(raw_list).__name__}"
            )
        template = raw_list[0]
        if console:
            console(
                f"Detected file template name {cls.template_pattern!r} in {module.relative_path.as_posix()!r}"
                f"Expanding file metadata..."
            )
        expanded_metadata: list[dict[str, Any]] = []
        for filepath in module.source_paths_by_resource_folder[FileLoader.folder_name]:
            if not FileLoader.is_supported_file(filepath):
                continue
            new_entry = copy.deepcopy(template)
            new_entry["externalId"] = new_entry["externalId"].replace(cls.template_pattern, filepath.name)
            new_entry["name"] = filepath.name
            expanded_metadata.append(new_entry)
        return expanded_metadata



from concurrent.futures import ThreadPoolExecutor
from itertools import islice
from datetime import timedelta
from typing import Any, Dict

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import NodeId, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteAsset, CogniteTimeSeries, CogniteTimeSeriesApply
from cognite.client.data_classes.filters import Prefix, ContainsAny
from cognite.client.exceptions import CogniteNotFoundError

import numpy as np

from cognite.client.config import global_config
global_config.disable_pypi_version_check = True


def batcher(iterable, batch_size):
    iterator = iter(iterable)
    while batch := list(islice(iterator, batch_size)):
        yield batch


def get_time_series_for_site(client: CogniteClient, site, space):
    this_site = site.lower()
    sub_tree_root = client.data_modeling.instances.retrieve_nodes(
        NodeId(space, this_site),
        node_cls=CogniteAsset
    )

    if not sub_tree_root:
        print(
            f"----No CogniteAssets in CDF for {site}!----\n"
            f"    Run the 'Create Cognite Asset Hierarchy' transformation!"
        )
        return

    sub_tree_nodes = client.data_modeling.instances.list(
        instance_type=CogniteAsset,
        filter=Prefix(property=["cdf_cdm", "CogniteAsset/v1", "path"], value=sub_tree_root.path),
        limit=None
    )

    if not sub_tree_nodes:
        print(
            f"----No CogniteTimeSeries in CDF for {site}!----\n"
            f"    Run the 'Contextualize Timeseries and Assets' transformation!"
        )
        return

    value_list = [{"space": node.space, "externalId": node.external_id} for node in sub_tree_nodes]

    time_series = [
        client.data_modeling.instances.search(
            view=ViewId("cdf_cdm", "CogniteTimeSeries", "v1"),
            instance_type=CogniteTimeSeries,
            space=space,
            filter=ContainsAny(property=["cdf_cdm", "CogniteTimeSeries/v1", "assets"], values=batch),
            limit=None
        )
        for batch in batcher(value_list, 20)
    ]

    # Combine list of batch results into a single NodeList
    time_series = [node for nodelist in time_series for node in nodelist]

    if not time_series:
        print("No CogniteTimeSeries in the CogniteCore Data Model (cdf_cdm Space)")

    return time_series

def handle(client: CogniteClient, data: Dict[str, Any] = {}) -> None:
    lookback_minutes = None
    sites = None

    if data:
        lookback_minutes = timedelta(minutes=data.get("lookback_minutes", 60)).total_seconds() * 1000
        sites = data.get("sites")

    all_sites = [
        "Houston",
        "Oslo",
        "Kuala_Lumpur",
        "Hannover",
        "Nuremberg",
        "Marseille",
        "Sao_Paulo",
        "Chicago",
        "Rotterdam",
        "London",
    ]

    lookback_minutes = lookback_minutes or timedelta(minutes=60).total_seconds() * 1000
    sites = sites or all_sites

    print(f"Processing datapoints for these sites: {sites}")
    with ThreadPoolExecutor(max_workers=10) as executor:
        futures = [executor.submit(process_site, client, lookback_minutes, site) for site in sites]
        for f in futures:
            f.result()

def process_site(client, lookback_minutes, site):
    oee_space = "oee_ts_space"
    source_space = "icapi_dm_space"

    timeseries = get_time_series_for_site(client, site, source_space)
    asset_eids = list(set([item.external_id.split(sep=":")[0] for item in timeseries]))
    instance_ids = [NodeId(space=source_space, external_id=ts.external_id) for ts in timeseries]
    all_latest_dps = client.time_series.data.retrieve_latest(instance_id=instance_ids)

    # Organize latest datapoints by equipment for alignment
    assets_dps = {
        external_id: [latest_dp for latest_dp in all_latest_dps if external_id in latest_dp.instance_id.external_id]
        for external_id in asset_eids
    }

    for asset, latest_dps in assets_dps.items():
        print(f"Calculating OEE for {asset}")
        count_node = f"NodeId({source_space}, {asset}:count)"
        good_node = f"NodeId({source_space}, {asset}:good)"
        status_node = f"NodeId({source_space}, {asset}:status)"
        planned_status_node = f"NodeId({source_space}, {asset}:planned_status)"

        end = min([dp.timestamp[0] for dp in latest_dps if latest_dps and dp.timestamp], default=None)

        if end:
            dps_df = client.time_series.data.retrieve_dataframe(
                instance_id=[dp.instance_id for dp in latest_dps],
                start=end - lookback_minutes,
                end=end,
                aggregates=["sum"],
                granularity="1m",
                include_aggregate_name=False,
                limit=None
            )

            # Frontfill because "planned_status" and "status" only have datapoints when the value changes
            dps_df = dps_df.ffill()

            # Fill the rest with the opposite
            try:
                first_valid_value = dps_df[planned_status_node].loc[dps_df[planned_status_node].first_valid_index()]
            except Exception as e:
                print(f"Failed to find datapoints for {planned_status_node}, {e}")
                continue

            backfill_value = 1.0 if first_valid_value == 0.0 else 0.0
            dps_df[planned_status_node] = dps_df[planned_status_node].fillna(value=backfill_value)

            # Same for status
            first_valid_value = dps_df[status_node].loc[dps_df[status_node].first_valid_index()]
            backfill_value = 1.0 if first_valid_value == 0.0 else 0.0
            dps_df[status_node] = dps_df[status_node].fillna(value=backfill_value)

            count_dps = dps_df[count_node]
            good_dps = dps_df[good_node]
            status_dps = dps_df[status_node]
            planned_status_dps = dps_df[planned_status_node]

            total_items = len(count_dps)

            if (
                total_items != len(good_dps)
                or total_items != len(status_dps)
                or total_items != len(planned_status_dps)
            ):
                # We expect ALL dependent timeseries to have the exact same number of datapoints
                # for the specified time range for the calculation to execute.
                print(
                    f"""{asset}: Unable to retrieve datapoints for all required OEE timeseries (count, good, status, planned_status)
                    between {end - lookback_minutes} and {end}. Ensure that data is available for the time range specified."""
                )

            # Calculate the components of OEE
            off_spec_node = f"{asset}:off_spec"
            quality_node = f"{asset}:quality"
            performance_node = f"{asset}:performance"
            availability_node = f"{asset}:availability"
            oee_node = f"{asset}:oee"

            dps_df[off_spec_node] = count_dps - good_dps
            dps_df[quality_node] = good_dps / count_dps
            dps_df[performance_node] = (count_dps / status_dps) / (60.0 / 3.0)
            dps_df[availability_node] = status_dps / planned_status_dps

            dps_df[oee_node] = dps_df[quality_node] * dps_df[performance_node] * dps_df[availability_node]

            # Fill in the divide by zeros
            dps_df = dps_df.fillna(value=0.0)
            dps_df = dps_df.replace([np.inf, -np.inf], 0.0)

            # Drop input timeseries
            dps_df = dps_df.drop(columns=[count_node, good_node, status_node, planned_status_node])

            to_insert = [
                {
                    "instance_id": NodeId(space="oee_ts_space", external_id=external_id),
                    "datapoints": list(zip(dps_df[external_id].index, dps_df[external_id]))
                }
                for external_id in dps_df.columns
            ]

            try:
                client.time_series.data.insert_multiple(to_insert)
            except CogniteNotFoundError as e:
                # Create the missing oee timeseries since they don't exist
                ts_to_create = []
                for node_id in e.not_found:
                    print(f"Creating CogniteTimeSeries {node_id}")

                    external_id = node_id["instanceId"]["externalId"]

                    # change external_id to a readable name
                    # Ex: "OSLPROFILTRASYS185:off_spec" to "OSLPROFILTRASYS185 Off Spec"
                    name = external_id.split(":")
                    name[-1] = name[-1].replace("_", " ").title()

                    ts_to_create.append(
                        CogniteTimeSeriesApply(
                            space=oee_space,
                            external_id=external_id,
                            name=" ".join(name),
                            is_step=False,
                            time_series_type="numeric",
                        )
                    )

                client.data_modeling.instances.apply(ts_to_create)
                client.time_series.data.insert_multiple(to_insert)



from __future__ import annotations

from cognite.client import ClientConfig, CogniteClient
from cognite.client._api_client import APIClient

from .agents import AgentsAPI


class AtlasAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.agents = AgentsAPI(config, api_version, cognite_client)



import hashlib
import json
from pathlib import Path
from typing import Any


def calculate_directory_hash(
    directory: Path,
    exclude_prefixes: set[str] | None = None,
    ignore_files: set[str] | None = None,
    shorten: bool = False,
) -> str:
    sha256_hash = hashlib.sha256()

    # Walk through each file in the directory
    for filepath in sorted(directory.rglob("*"), key=lambda p: str(p.relative_to(directory))):
        if filepath.is_dir():
            continue
        if exclude_prefixes and any(filepath.name.startswith(prefix) for prefix in exclude_prefixes):
            continue
        if ignore_files and filepath.suffix in ignore_files:
            continue
        relative_path = filepath.relative_to(directory)
        sha256_hash.update(relative_path.as_posix().encode("utf-8"))
        # Open each file and update the hash
        with filepath.open("rb") as file:
            while chunk := file.read(8192):
                # Get rid of Windows line endings to make the hash consistent across platforms.
                sha256_hash.update(chunk.replace(b"\r\n", b"\n"))

    calculated = sha256_hash.hexdigest()
    if shorten:
        return calculated[:8]
    return calculated


def calculate_secure_hash(item: dict[str, Any], shorten: bool = False) -> str:
    """Calculate a secure hash of a dictionary"""
    sha256_hash = hashlib.sha512(usedforsecurity=True)
    sha256_hash.update(json.dumps(item, sort_keys=True).encode("utf-8"))
    calculated_hash = sha256_hash.hexdigest()
    if shorten:
        return calculated_hash[:8]
    return calculated_hash


def calculate_hash(content: str | bytes | Path, shorten: bool = False) -> str:
    sha256_hash = hashlib.sha256()
    if isinstance(content, Path):
        # Get rid of Windows line endings to make the hash consistent across platforms.
        content = content.read_bytes().replace(b"\r\n", b"\n")
    elif isinstance(content, str):
        content = content.encode("utf-8")
    sha256_hash.update(content)
    calculated = sha256_hash.hexdigest()
    if shorten:
        return calculated[:8]
    return calculated



from __future__ import annotations

from collections.abc import Iterator
from typing import overload

from cognite.client import CogniteClient
from cognite.client._api_client import APIClient
from cognite.client.config import ClientConfig

from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    LocationFilter,
    LocationFilterList,
    LocationFilterWrite,
)


class LocationFiltersAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        # this is hopefully a very temporary fix to avoid the client concatinating the _RESOURCE_PATH with the ordinary /api/ path.
        self._api_version = None
        self._RESOURCE_PATH = f"/apps/v1/projects/{self._cognite_client.config.project}/storage/config/locationfilters"

    @overload
    def __call__(self) -> Iterator[LocationFilter]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[LocationFilterList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[LocationFilter] | Iterator[LocationFilterList]:
        """Iterate over locationfilters.

        Args:
            chunk_size: The number of locationfilters to return in each chunk. None will return all locationfilters.

        Yields:
            LocationFilter or LocationFilterList

        """
        return iter(self.list())

    def __iter__(self) -> Iterator[LocationFilter]:
        return self.__call__()

    def create(self, location_filter: LocationFilterWrite) -> LocationFilter:
        """Create a new LocationFilter.

        Args:
            location_filter: LocationFilterWrite


        Returns:
            LocationFilter

        """

        res = self._post(url_path=self._RESOURCE_PATH, json=location_filter.dump())
        return LocationFilter._load(res.json(), cognite_client=self._cognite_client)

    def retrieve(self, id: int) -> LocationFilter:
        """Retrieve a single LocationFilter.

        Args:
            id: The ID of the LocationFilter

        Returns:
            LocationFilter
        """

        res = self._get(url_path=f"{self._RESOURCE_PATH}/{id}")
        return LocationFilter._load(res.json(), cognite_client=self._cognite_client)

    def update(self, id: int, location_filter_update: LocationFilterWrite) -> LocationFilter:
        """Update a new LocationFilter.

        Args:
            id: The ID of the LocationFilter
            location_filter_content: The content of the LocationFilter to update

        Returns:
            LocationFilter

        """

        res = self._put(
            url_path=f"{self._RESOURCE_PATH}/{id}",
            json=location_filter_update.dump(),
        )
        return LocationFilter._load(res.json(), cognite_client=self._cognite_client)

    def delete(self, id: int) -> None:
        self._delete(url_path=f"{self._RESOURCE_PATH}/{id}")

    def list(self) -> LocationFilterList:
        res = self._post(url_path=self._RESOURCE_PATH + "/list", json={"flat": True})
        return LocationFilterList._load(res.json()["items"], cognite_client=self._cognite_client)



from __future__ import annotations

from enum import Enum
from functools import lru_cache
from typing import Any, ClassVar

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import clean_name


class Flags(Enum):
    # Ruff requires annotations while mypy requires no annotations
    INTERNAL: ClassVar[dict[str, Any]] = {"visible": False, "description": "Does nothing"}  # type: ignore[misc]
    IMPORT_CMD: ClassVar[dict[str, Any]] = {"visible": True, "description": "Enables the import sub application"}  # type: ignore[misc]
    GRAPHQL: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables the support for deploying data models as GraphQL schemas",
    }
    MODULE_REPEAT: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables the support for repeating modules in the config file",
    }
    POPULATE: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables support for the populate command",
    }
    AGENTS: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables support for Atlas Agents and Agent Tools",
    }
    DUMP_EXTENDED: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables support for the dumping Location Filters.",
    }
    PROFILE: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables support for the profile command",
    }
    INFIELD: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables support for Infield configs",
    }
    DUMP_DATA: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Splits the dump command in dump data and dump config",
    }
    EXIT_ON_WARNING: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables the exit on warning feature during the build command",
    }
    MIGRATE: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "visible": True,
        "description": "Enables the migrate command",
    }

    def is_enabled(self) -> bool:
        return FeatureFlag.is_enabled(self)


class FeatureFlag:
    @staticmethod
    @lru_cache(typed=True)
    def is_enabled(flag: Flags) -> bool:
        return CDFToml.load().alpha_flags.get(clean_name(flag.name), False)



from collections.abc import Iterable
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.commands.pull import PullCommand, ResourceYAMLDifference, TextFileDifference
from cognite_toolkit._cdf_tk.data_classes import (
    BuildVariable,
    BuildVariables,
    BuiltFullResourceList,
    BuiltResourceFull,
)
from cognite_toolkit._cdf_tk.loaders import DataSetsLoader
from tests.test_unit.approval_client import ApprovalToolkitClient


def load_update_diffs_use_cases():
    build_file = """externalId: tr_pump_asset_hierarchy-load-collections_pump
name: pump_asset_hierarchy-load-collections_pump
destination:
  type: asset_hierarchy
dataSetExternalId: src:lift_pump_stations
ignoreNullFields: false
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: ${IDP_CLIENT_ID}
  clientSecret: ${IDP_CLIENT_SECRET}
  tokenUri: ${IDP_TOKEN_URL}
  # Optional: If idP requires providing the scopes
  cdfProjectName: ${CDF_PROJECT}
  scopes:
  - ${IDP_SCOPES}
  # Optional: If idP requires providing the audience
  audience: ${IDP_AUDIENCE}
"""
    source_file = """externalId: tr_pump_asset_hierarchy-load-collections_pump
name: pump_asset_hierarchy-load-collections_pump
destination:
  type: asset_hierarchy
dataSetExternalId: {{data_set}}
ignoreNullFields: false
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: {{cicd_clientId}}
  clientSecret: {{cicd_clientSecret}}
  tokenUri: {{cicd_tokenUri}}
  # Optional: If idP requires providing the scopes
  cdfProjectName: {{cdfProjectName}}
  scopes: {{cicd_scopes}}
  # Optional: If idP requires providing the audience
  audience: {{cicd_audience}}
"""
    cdf_resource = {
        "conflictMode": "upsert",
        "destination": {"type": "asset_hierarchy"},
        "externalId": "tr_pump_asset_hierarchy-load-collections_pump",
        "ignoreNullFields": False,
        "isPublic": True,
        "name": "pump_asset_hierarchy-load-collections_pump",
    }
    expected = {
        "added": {"isPublic": True, "conflictMode": "upsert"},
        "changed": {},
        "cannot_change": {},
    }

    dumped = """externalId: tr_pump_asset_hierarchy-load-collections_pump
name: pump_asset_hierarchy-load-collections_pump
destination:
  type: asset_hierarchy
dataSetExternalId: {{data_set}}
ignoreNullFields: false
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: {{cicd_clientId}}
  clientSecret: {{cicd_clientSecret}}
  tokenUri: {{cicd_tokenUri}}
  # Optional: If idP requires providing the scopes
  cdfProjectName: {{cdfProjectName}}
  scopes: {{cicd_scopes}}
  # Optional: If idP requires providing the audience
  audience: {{cicd_audience}}
conflictMode: upsert
isPublic: true
"""

    yield pytest.param(build_file, source_file, cdf_resource, expected, dumped, id="Transformation with no differences")

    build_file = """autoCreateDirectRelations: true
skipOnVersionConflict: false
replace: true
nodes:
- space: APM_Config
  externalId: default_infield_config_minimal
  sources:
  - source:
      space: APM_Config
      externalId: APM_Config
      version: '1'
      type: view
    properties:
      featureConfiguration:
        rootLocationConfigurations:
        - assetExternalId: WMT:VAL
          appDataInstanceSpace: sp_infield_oid_app_data
          sourceDataInstanceSpace: sp_asset_oid_source
          templateAdmins:
          - gp_infield_oid_template_admins
          checklistAdmins:
          - gp_infield_oid_checklist_admins
      customerDataSpaceId: APM_SourceData
      customerDataSpaceVersion: '1'
      name: Default location
"""
    source_file = """autoCreateDirectRelations: true
skipOnVersionConflict: false
replace: true
nodes:
- space: {{apm_config_instance_space}}
  externalId: default_infield_config_minimal
  sources:
  - source:
      space: APM_Config
      externalId: APM_Config
      version: '1'
      type: view
    properties:
      featureConfiguration:
        rootLocationConfigurations:
        - assetExternalId: {{root_asset_external_id}}
          appDataInstanceSpace: sp_infield_{{default_location}}_app_data
          sourceDataInstanceSpace: sp_asset_{{default_location}}_source
          templateAdmins:
          - gp_infield_{{default_location}}_template_admins
          checklistAdmins:
          - gp_infield_{{default_location}}_checklist_admins
      customerDataSpaceId: APM_SourceData
      customerDataSpaceVersion: '1'
      name: Default location
"""

    cdf_resource = {
        "autoCreateDirectRelations": True,
        "skipOnVersionConflict": False,
        "replace": True,
        "nodes": [
            {
                "space": "APM_Config",
                "instanceType": "node",
                "externalId": "default_infield_config_minimal",
                "sources": [
                    {
                        "properties": {
                            "name": "Default location",
                            "appDataSpaceVersion": "1",
                            "customerDataSpaceId": "APM_SourceData",
                            "featureConfiguration": {
                                "rootLocationConfigurations": [
                                    {
                                        "templateAdmins": ["gp_infield_oid_template_admins"],
                                        "assetExternalId": "WMT:VAL",
                                        "checklistAdmins": ["gp_infield_oid_checklist_admins"],
                                        "appDataInstanceSpace": "sp_infield_oid_app_data",
                                        "sourceDataInstanceSpace": "sp_asset_oid_source",
                                    }
                                ]
                            },
                            "customerDataSpaceVersion": "1",
                        },
                        "source": {"space": "APM_Config", "externalId": "APM_Config", "version": "1", "type": "view"},
                    }
                ],
            }
        ],
    }

    expected = {
        "added": {
            "nodes.0.instanceType": "node",
            "nodes.0.sources.0.properties.appDataSpaceVersion": "1",
        },
        "changed": {},
        "cannot_change": {},
    }

    dumped = """autoCreateDirectRelations: true
skipOnVersionConflict: false
replace: true
nodes:
- space: {{apm_config_instance_space}}
  externalId: default_infield_config_minimal
  sources:
  - source:
      space: APM_Config
      externalId: APM_Config
      version: '1'
      type: view
    properties:
      featureConfiguration:
        rootLocationConfigurations:
        - assetExternalId: {{root_asset_external_id}}
          appDataInstanceSpace: sp_infield_{{default_location}}_app_data
          sourceDataInstanceSpace: sp_asset_{{default_location}}_source
          templateAdmins:
          - gp_infield_{{default_location}}_template_admins
          checklistAdmins:
          - gp_infield_{{default_location}}_checklist_admins
      customerDataSpaceId: APM_SourceData
      customerDataSpaceVersion: '1'
      name: Default location
      appDataSpaceVersion: '1'
  instanceType: node
"""

    yield pytest.param(build_file, source_file, cdf_resource, expected, dumped, id="Node with a changed field")

    build_file = """externalId: tr_timeseries_oid_pi_apm_simple_load_timeseries2assets
name: timeseries:oid:pi:apm_simple:load_timeseries2assets
destination:
  dataModel:
    space: sp_apm_simple
    externalId: apm_simple
    version: '1'
    destinationType: Asset
  instanceSpace: sp_apm_simple
  type: instances
ignoreNullFields: true
shared: true
action: upsert
dataSetExternalId: ds_transformations_oid
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: ${IDP_CLIENT_ID}
  clientSecret: ${IDP_CLIENT_SECRET}
  tokenUri: ${IDP_TOKEN_URL}
  # Optional: If idP requires providing the cicd_scopes
  cdfProjectName: ${CDF_PROJECT}
  scopes: ['${IDP_SCOPES}']
  # Optional: If idP requires providing the cicd_audience
  audience: ${IDP_AUDIENCE}"""

    source_file = """externalId: tr_timeseries_{{default_location}}_{{source_timeseries}}_apm_simple_load_timeseries2assets
name: timeseries:{{default_location}}:{{source_timeseries}}:apm_simple:load_timeseries2assets
destination:
  dataModel:
    space: {{space}}
    externalId: {{datamodel}}
    version: '1'
    destinationType: Asset
  instanceSpace: {{space}}
  type: instances
ignoreNullFields: true
shared: true
action: upsert
dataSetExternalId: ds_transformations_{{default_location}}
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: {{cicd_clientId}}
  clientSecret: {{cicd_clientSecret}}
  tokenUri: {{cicd_tokenUri}}
  # Optional: If idP requires providing the cicd_scopes
  cdfProjectName: {{cdfProjectName}}
  scopes: {{cicd_scopes}}
  # Optional: If idP requires providing the cicd_audience
  audience: {{cicd_audience}}"""

    cdf_resource = {
        "externalId": "tr_timeseries_oid_pi_apm_simple_load_timeseries2assets",
        "name": "timeseries:oid:pi:apm_simple:load_timeseries2assets",
        "destination": {
            "type": "instances",
            "dataModel": {
                "space": "sp_apm_simple",
                "externalId": "apm_simple",
                "version": "1",
                "destinationType": "Asset",
                "destinationRelationshipFromType": None,
            },
            "instanceSpace": "sp_apm_simple",
        },
        "conflictMode": "upsert",
        "isPublic": True,
        "ignoreNullFields": True,
    }

    expected = {
        "added": {"conflictMode": "upsert", "isPublic": True},
        "changed": {},
        "cannot_change": {},
    }

    dumped = """externalId: tr_timeseries_{{default_location}}_{{source_timeseries}}_apm_simple_load_timeseries2assets
name: timeseries:{{default_location}}:{{source_timeseries}}:apm_simple:load_timeseries2assets
destination:
  dataModel:
    space: {{space}}
    externalId: {{datamodel}}
    version: '1'
    destinationType: Asset
    destinationRelationshipFromType: null
  instanceSpace: {{space}}
  type: instances
ignoreNullFields: true
shared: true
action: upsert
dataSetExternalId: ds_transformations_{{default_location}}
# Specify credentials separately like this:
# You can also use different credentials for the running transformations than the ones you use to deploy
authentication:
  clientId: {{cicd_clientId}}
  clientSecret: {{cicd_clientSecret}}
  tokenUri: {{cicd_tokenUri}}
  # Optional: If idP requires providing the cicd_scopes
  cdfProjectName: {{cdfProjectName}}
  scopes: {{cicd_scopes}}
  # Optional: If idP requires providing the cicd_audience
  audience: {{cicd_audience}}
conflictMode: upsert
isPublic: true
"""

    yield pytest.param(build_file, source_file, cdf_resource, expected, dumped, id="Transformation double variable.")


class TestResourceYAML:
    @pytest.mark.parametrize(
        "build_file, source_file, cdf_resource, expected, expected_dumped",
        list(load_update_diffs_use_cases()),
    )
    def test_load_update_changes_dump(
        self,
        build_file: str,
        source_file: str,
        cdf_resource: dict[str, Any],
        expected: dict[str, dict[str, Any]],
        expected_dumped: str,
    ) -> None:
        resource_yaml = ResourceYAMLDifference.load(build_file, source_file)
        resource_yaml.update_cdf_resource(cdf_resource)

        added = {".".join(map(str, key)): value.cdf_value for key, value in resource_yaml.items() if value.is_added}
        changed = {".".join(map(str, key)): value.cdf_value for key, value in resource_yaml.items() if value.is_changed}
        cannot_change = {
            ".".join(map(str, key)): value.cdf_value for key, value in resource_yaml.items() if value.is_cannot_change
        }

        assert added == expected["added"]
        assert changed == expected["changed"]
        assert cannot_change == expected["cannot_change"]

        assert resource_yaml.dump_yaml_with_comments() == expected_dumped


def load_update_changed_dump_test_cases():
    build_file = """--- 1. asset root (defining all columns)
SELECT
    "Lift Pump Stations" AS name,
    dataset_id("src:lift_pump_stations") AS dataSetId,
    "lift_pump_stations:root" AS externalId,
    '' as parentExternalId,
    "An example pump dataset" as description,
    null as metadata
"""
    source_file = """--- 1. asset root (defining all columns)
SELECT
    "Lift Pump Stations" AS name,
    dataset_id("{{data_set}}") AS dataSetId,
    "lift_pump_stations:root" AS externalId,
    '' as parentExternalId,
    "An example pump dataset" as description,
    null as metadata
"""
    cdf_content = """--- 1. asset root (defining all columns) And Extra comment in the SQL
SELECT
    "Lift Pump Stations" AS name,
    dataset_id("src:new_data_set") AS dataSetId,
    "lift_pump_stations:root" AS externalId,
    '' as parentExternalId,
    "An example pump dataset" as description,
    null as metadata
"""
    expected = {
        "added": [],
        "changed": ["--- 1. asset root (defining all columns) And Extra comment in the SQL"],
        "cannot_change": [('    dataset_id("src:new_data_set") AS dataSetId,', ["data_set"])],
    }
    dumped = """--- 1. asset root (defining all columns) And Extra comment in the SQL
SELECT
    "Lift Pump Stations" AS name,
    dataset_id("{{data_set}}") AS dataSetId,
    "lift_pump_stations:root" AS externalId,
    '' as parentExternalId,
    "An example pump dataset" as description,
    null as metadata
"""
    yield pytest.param(build_file, source_file, cdf_content, expected, dumped, id="SQL with one line differences")


class TestTextFileDifference:
    @pytest.mark.parametrize(
        "build_file, source_file, cdf_content, expected, dumped",
        list(load_update_changed_dump_test_cases()),
    )
    def test_load_update_changes_dump(
        self, build_file: str, source_file: str, cdf_content: str, expected: dict[str, list], dumped: str
    ) -> None:
        text_file = TextFileDifference.load(build_file, source_file)
        text_file.update_cdf_content(cdf_content)

        added = [line.cdf_value for line in text_file if line.is_added]
        changed = [line.cdf_value for line in text_file if line.is_changed]
        cannot_change = [(line.cdf_value, line.variables) for line in text_file if line.is_cannot_change]

        assert added == expected["added"]
        assert changed == expected["changed"]
        assert cannot_change == expected["cannot_change"]

        assert text_file.dump() == dumped


def to_write_content_use_cases() -> Iterable:
    source = """name: Ingestion
externalId: {{ dataset }}
description: This dataset contains Transformations, Functions, and Workflows for ingesting data into Cognite Data Fusion.
"""
    to_write = {"ingestion": {"name": "Ingestion", "externalId": "ingestion", "description": "New description"}}
    variable = BuildVariable(
        key="dataset",
        value="ingestion",
        is_selected=True,
        location=Path("whatever"),
    )
    ingestion = MagicMock(spec=BuiltResourceFull)
    ingestion.build_variables = BuildVariables([variable])
    ingestion.identifier = "ingestion"
    ingestion.extra_sources = []

    resources = BuiltFullResourceList([ingestion])

    expected = """name: Ingestion
externalId: {{ dataset }}
description: New description
"""

    yield pytest.param(source, to_write, resources, expected, id="One resource changed")

    source = """name: Ingestion
externalId: {{ dataset }} # This is a comment
# This is another comment
description: Original description
"""

    expected = """name: Ingestion
externalId: {{ dataset }} # This is a comment
# This is another comment
description: New description
"""

    yield pytest.param(source, to_write, resources, expected, id="One resource changed with comments")

    source = """- name: Ingestion
  externalId: {{ dataset }} # This is a comment
  # This is another comment
  description: Original description
- name: Another
  externalId: unique_dataset
  description: with its own description
"""

    expected = """- name: Ingestion
  externalId: {{ dataset }} # This is a comment
  # This is another comment
  description: New description
- name: Another
  externalId: unique_dataset
  description: also new description
"""
    to_write_multi = {
        **to_write,
        "unique_dataset": {"name": "Another", "externalId": "unique_dataset", "description": "also new description"},
    }
    unique_dataset = MagicMock(spec=BuiltResourceFull)
    unique_dataset.build_variables = BuildVariables([])
    unique_dataset.identifier = "unique_dataset"
    unique_dataset.extra_sources = []
    resources = BuiltFullResourceList([ingestion, unique_dataset])

    yield pytest.param(source, to_write_multi, resources, expected, id="Multiple resources changed")


class TestPullCommand:
    @pytest.mark.parametrize(
        "source, to_write, resources, expected",
        list(to_write_content_use_cases()),
    )
    def test_to_write_content(
        self,
        source: str,
        to_write: dict[str, [dict[str, Any]]],
        resources: BuiltFullResourceList,
        expected: str,
        toolkit_client_approval: ApprovalToolkitClient,
    ) -> None:
        cmd = PullCommand(silent=True, skip_tracking=True)

        actual, extra_files = cmd._to_write_content(
            source=source,
            to_write=to_write,
            resources=resources,
            environment_variables={},
            loader=DataSetsLoader.create_loader(toolkit_client_approval.mock_client),
        )
        assert not extra_files, "This tests does not support testing extra files"
        assert actual.splitlines() == expected.splitlines()



"""Contextualize handler."""

import logging
import sys

from cognite.client.data_classes import FileMetadataUpdate, TimeSeries
from common.cdf_helpers import create_missing_labels

MAX_GET_JOB_ATTEMPTS = 10
logger = logging.getLogger(__name__)


def get_time_series_label(camera_type):
    if camera_type == "ir":
        time_series_label = "IR_TIME_SERIES"
    else:
        time_series_label = "ANALOG_GAUGE_TIME_SERIES"
    return time_series_label


def has_label(label_list, label_external_id):
    """Check if the label_external_id is in a list of labels."""
    for label in label_list:
        if label.external_id == label_external_id:
            return True
    return False


def create_timeseries(
    client,
    asset,
    data_set_external_id,
    time_series_label,
    gauge_number=1,
    total_number_of_gauges=1,
    time_series_naming_tag="",
):
    """Create time series on the correct format."""
    if time_series_label == "IR_TIME_SERIES":
        ts_external_id = (
            f"{time_series_label}_{asset.external_id}_{time_series_naming_tag}"
            if asset.external_id is not None
            else f"{asset.id}_{time_series_label}_{time_series_naming_tag}"
        )
        ts_name = f"{asset.name}_ir_{time_series_naming_tag}"

    elif gauge_number <= 1 and total_number_of_gauges <= 1:
        ts_external_id = (
            f"{time_series_label}_{asset.external_id}_0"
            if asset.external_id is not None
            else f"{asset.id}_{time_series_label}_0"
        )
        ts_name = f"{asset.name}_gauge_values"
    else:
        ts_external_id = (
            f"{time_series_label}_{asset.external_id}_{gauge_number}"
            if asset.external_id is not None
            else f"{asset.id}_{time_series_label}_{gauge_number}"
        )
        ts_name = f"{asset.name}_gauge_values_{gauge_number}"

    logger.info(f"Creates timeseries with name: {ts_name}")
    data_set_id = client.data_sets.retrieve(external_id=data_set_external_id).id
    return client.time_series.create(
        TimeSeries(
            metadata={"TYPE": time_series_label},
            asset_id=asset.id,
            external_id=ts_external_id,
            data_set_id=data_set_id,
            name=ts_name,
        )
    )


def get_asset(client, file):
    """Get asset from metadata field."""
    if file.metadata.get("asset_id") not in [None, "0", "None"]:
        try:
            asset_id = int(file.metadata["asset_id"])
            asset = client.assets.retrieve(asset_id)
        except Exception as e:
            logger.exception(e)
            return None
        return asset


def get_timeseries_external_id(client, asset, data_set_external_id, time_series_label, number_of_gauges=1):
    """Get time series external id."""
    ts = client.time_series.list(metadata={"TYPE": time_series_label}, asset_ids=[asset.id])
    number_of_timeseries = len(ts)
    ts_external_ids = []
    logger.info(f"{ts}")

    # Create one timeseries for min temperature and one for max temperature for IR images
    if time_series_label == "IR_TIME_SERIES":
        if number_of_timeseries == 0:
            print("CREATE IR TIMESERIES")
            ir_ts_naming_tags = ["min", "max"]
            for naming_tag in ir_ts_naming_tags:
                ts = create_timeseries(
                    client,
                    asset,
                    data_set_external_id,
                    time_series_label,
                    time_series_naming_tag=naming_tag,
                )
                ts_external_ids.append(ts.external_id)
        else:
            ts_external_ids = [ts[i].external_id for i in range(number_of_timeseries)]

    # Logic for creating gauge timeseries
    elif number_of_timeseries >= 1 and number_of_gauges == number_of_timeseries:
        ts_external_ids = [ts[i].external_id for i in range(number_of_timeseries)]
    elif number_of_timeseries >= 1 and number_of_gauges > number_of_timeseries:
        ts_external_ids.append(ts[0].external_id)
        diff = number_of_gauges - number_of_timeseries
        for gauge_number in range(diff):
            logger.info("CREATE GAUGE TIMESERIES")
            new_gauge_number = gauge_number + number_of_timeseries
            ts = create_timeseries(
                client, asset, data_set_external_id, time_series_label, new_gauge_number, number_of_gauges
            )
            ts_external_ids.append(ts.external_id)
    elif number_of_timeseries == 0:
        logger.info("CREATE GAUGE TIMESERIES")
        if number_of_gauges <= 1:
            ts = create_timeseries(client, asset, data_set_external_id, time_series_label)
            ts_external_ids.append(ts.external_id)
        else:
            for gauge_number in range(number_of_gauges):
                ts = create_timeseries(
                    client, asset, data_set_external_id, time_series_label, gauge_number, number_of_gauges
                )
                ts_external_ids.append(ts.external_id)

    return ts_external_ids


def get_number_of_gauges(client, asset):
    """Get the total number of gauges associated with an asset."""
    try:
        asset_data = client.assets.retrieve(id=asset.id)
        asset_metadata = asset_data.metadata
        if "number_of_gauges" in asset_metadata:
            return int(asset_metadata["number_of_gauges"])
        else:
            return 1
    except Exception as e:
        logger.exception(e)
        return 1


def handle(data, client):
    """Contextualize robot data."""
    print("Start gauge reading.")
    required_input_data_fields = {
        "data_set_external_id",
        "read_dial_gauge_label",
        "read_multiple_dial_gauges_label",
        "read_digital_gauge_label",
        "read_level_gauge_label",
        "read_valve_label",
        "read_ir_raw_label",
        "spill_detection_label",
        "gauge_context_label",
    }
    if not required_input_data_fields <= data.keys():
        raise RuntimeError(f"Data should contain all keys:  {required_input_data_fields}. Current data: {data}")

    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=logging.BASIC_FORMAT)
    create_missing_labels(
        client,
        [
            data["read_dial_gauge_label"],
            data["read_multiple_dial_gauges_label"],
            data["read_digital_gauge_label"],
            data["read_level_gauge_label"],
            data["read_valve_label"],
            data["read_ir_raw_label"],
            data["spill_detection_label"],
            "threesixty",
        ],
    )
    files = client.files.list(
        data_set_external_ids=[data["data_set_external_id"]],
        metadata={"processed": "false"},
        limit=-1,
        uploaded=True,
    )

    logger.info(f"Number of files to process in dataset  {data['data_set_external_id']}: {len(files)}")
    print(f"Number of files to process in dataset  {data['data_set_external_id']}: {len(files)}")

    for file in files:
        number_of_gauges = 1  # Default is one gauge per image
        metadata = file.metadata
        logger.info(f"Processing file with external id: {file.external_id}, id: {file.id}")

        update = FileMetadataUpdate(file.id)
        asset = get_asset(client, file)
        update.asset_ids.add([asset.id]) if asset else None

        print(f"The labels on this file are: {file.labels}")
        if file.labels and has_label(file.labels, data["gauge_context_label"]):
            gauge_type = metadata.get("gauge_type", "dial")  # Default gauge type from robot is dial
            print(f"This file has gauge type: {gauge_type}")

            camera_type = metadata.get("camera", None)
            time_series_label = get_time_series_label(camera_type)

            if gauge_type == "valve":
                current_action_label = "read_valve_label"
            elif gauge_type == "dial":
                if asset:
                    number_of_gauges = get_number_of_gauges(client, asset)
                if number_of_gauges <= 1:
                    current_action_label = "read_dial_gauge_label"
                else:
                    current_action_label = "read_multiple_dial_gauges_label"
            else:
                current_action_label = f"read_{gauge_type}_gauge_label"

            logger.info(f"Setting action label to {data[current_action_label]}")
            print(f"Setting action label to {data[current_action_label]}")

            update.labels.add(data[current_action_label])
            update.labels.remove(data["gauge_context_label"])

            if asset and gauge_type != "spill":
                logger.info(f"Number of gauges: {number_of_gauges}.")
                ts_eid = get_timeseries_external_id(
                    client, asset, data["data_set_external_id"], time_series_label, number_of_gauges
                )
                logger.info(f"Time series linked to asset: {ts_eid}.")
                if ts_eid:
                    if time_series_label == "IR_TIME_SERIES":
                        for i in range(len(ts_eid)):
                            metadata[f"ts_external_id_{i}"] = ts_eid[i]
                    elif number_of_gauges <= 1:
                        metadata["ts_external_id"] = ts_eid[0]
                    else:
                        for gauge_number in range(number_of_gauges):
                            if metadata.get("ts_external_id", None):
                                if gauge_number == 0:
                                    continue
                                else:
                                    metadata[f"ts_external_id_{gauge_number}"] = ts_eid[gauge_number]
                            else:
                                metadata[f"ts_external_id_{gauge_number}"] = ts_eid[gauge_number]
        elif "threesixty" in file.external_id:
            print("This is a 360 image, adding threesixty label.")
            update.labels.add(["threesixty"])

        elif asset and metadata.get("method") == "process_ir_raw":
            current_action_label = "read_ir_raw_label"
            time_series_label = "IR_TIME_SERIES"
            print(f"this is the asset: {asset}")

            ts_eid = get_timeseries_external_id(
                client, asset, data["data_set_external_id"], time_series_label, number_of_gauges
            )
            if time_series_label == "IR_TIME_SERIES":
                for i in range(len(ts_eid)):
                    metadata[f"ts_external_id_{i}"] = ts_eid[i]

        metadata["processed"] = "true"
        update.metadata.add(metadata)
        client.files.update(update)
    return {}



import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING

from cognite.client.config import ClientConfig
from cognite.client.data_classes.capabilities import Capability
from cognite.client.data_classes.iam import TokenInspection
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client.api_client import ToolkitAPI
from cognite_toolkit._cdf_tk.constants import URL
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError

if TYPE_CHECKING:
    from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient


class VerifyAPI(ToolkitAPI):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: "ToolkitClient") -> None:
        super().__init__(config, api_version, cognite_client)
        self._token_inspect: TokenInspection | None = None

    @property
    def token_inspect(self) -> TokenInspection:
        if self._token_inspect is None:
            with warnings.catch_warnings():
                # If the user has unknown capabilities, we don't want the user to see the warning:
                # "UserWarning: Unknown capability '<unknown warning>'.
                warnings.simplefilter("ignore")
                try:
                    self._token_inspect = self._toolkit_client.iam.token.inspect()
                except CogniteAPIError as e:
                    raise AuthorizationError(
                        f"Don't seem to have any access rights. {e}\n"
                        f"Please visit [link={URL.configure_access}]the documentation[/link] "
                        f"and ensure you have configured your access correctly."
                    ) from e
        return self._token_inspect

    def authorization(self, capabilities: Capability | Sequence[Capability]) -> list[Capability]:
        """Verify that the client has correct credentials and required access rights

        Args:
            capabilities (Capability | Sequence[Capability]): access capabilities to verify

        Returns:
            list[Capability]: list of missing capabilities

        """
        with warnings.catch_warnings():
            # If the user has unknown capabilities, we don't want the user to see the warning:
            # "UserWarning: Unknown capability '<unknown warning>' will be ignored in comparison"
            # This is irrelevant for the user as we are only checking the capabilities that are known.
            warnings.simplefilter("ignore")
            return self._toolkit_client.iam.compare_capabilities(self.token_inspect.capabilities, capabilities)

    @staticmethod
    def create_error(missing_capabilities: Sequence[Capability], action: str | None = None) -> AuthorizationError:
        """Create an AuthorizationError with a message that lists the missing capabilities

        Args:
            missing_capabilities (Sequence[Capability]): capabilities that are missing
            action (str, optional): action that requires the capabilities. Defaults to None.

        """
        if not missing_capabilities:
            raise ValueError("Bug in Toolkit. Tried creating an AuthorizationError without any missing capabilities.")
        missing = "  - \n".join(repr(c) for c in missing_capabilities)
        first_sentence = "Don't have correct access rights"
        if action:
            first_sentence += f" to {action}."
        else:
            first_sentence += "."

        return AuthorizationError(
            f"{first_sentence} Missing:\n{missing}\n"
            f"Please [blue][link={URL.auth_toolkit}]click here[/link][/blue] to visit the documentation "
            "and ensure that you have setup authentication for the CDF toolkit correctly."
        )



import sys
from pathlib import Path
from typing import cast
from unittest.mock import MagicMock

import pytest
import yaml
from cognite.client import data_modeling as dm
from cognite.client.data_classes import (
    DataSet,
    Group,
    GroupWrite,
    Transformation,
    Workflow,
    WorkflowDefinition,
    WorkflowTrigger,
    WorkflowVersion,
    WorkflowVersionId,
)
from cognite.client.data_classes.capabilities import AssetsAcl, EventsAcl, TimeSeriesAcl
from cognite.client.data_classes.workflows import WorkflowScheduledTriggerRule
from pytest import MonkeyPatch

from cognite_toolkit._cdf_tk import cdf_toml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import LocationFilter
from cognite_toolkit._cdf_tk.commands import BuildCommand, DeployCommand, DumpResourceCommand, PullCommand
from cognite_toolkit._cdf_tk.commands.dump_resource import DataModelFinder, WorkflowFinder
from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.data_classes import BuildConfigYAML, Environment
from cognite_toolkit._cdf_tk.exceptions import ToolkitDuplicatedModuleError
from cognite_toolkit._cdf_tk.loaders import RESOURCE_LOADER_LIST
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import (
    BUILD_GROUP_WITH_UNKNOWN_ACL,
    COMPLETE_ORG_ONLY_IDENTIFIER,
    NAUGHTY_PROJECT,
    PROJECT_FOR_TEST,
    PROJECT_NO_COGNITE_MODULES,
    PROJECT_WITH_DUPLICATES,
)
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import mock_read_yaml_file


def test_inject_custom_environmental_variables(
    build_tmp_path: Path,
    monkeypatch: MonkeyPatch,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    organization_dir: Path,
) -> None:
    config_yaml = yaml.safe_load((organization_dir / "config.dev.yaml").read_text())
    config_yaml["variables"]["modules"]["cdf_common"]["dataset"] = "${MY_ENVIRONMENT_VARIABLE}"
    # Selecting the cdf_common module to be built
    config_yaml["environment"]["selected"] = ["cdf_common"]
    config_yaml["environment"]["project"] = "pytest"
    mock_read_yaml_file(
        {
            "config.dev.yaml": config_yaml,
        },
        monkeypatch,
    )
    monkeypatch.setenv("MY_ENVIRONMENT_VARIABLE", "my_environment_variable_value")
    BuildCommand(silent=True).execute(
        organization_dir=organization_dir,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=None,
        on_error="raise",
        verbose=False,
    )
    DeployCommand(silent=True).execute(
        env_vars=env_vars_with_client,
        build_dir=build_tmp_path,
        build_env_name="dev",
        drop=True,
        dry_run=False,
        include=[],
        drop_data=False,
        verbose=False,
        force_update=False,
    )

    dataset = toolkit_client_approval.created_resources_of_type(DataSet)[0]
    assert dataset.external_id == "my_environment_variable_value"


def test_duplicated_modules(build_tmp_path: Path) -> None:
    config = MagicMock(spec=BuildConfigYAML)
    config.environment = MagicMock(spec=Environment)
    config.environment.name = "dev"
    config.environment.selected = ["module1"]
    with pytest.raises(ToolkitDuplicatedModuleError) as err:
        BuildCommand().build_config(
            build_dir=build_tmp_path,
            organization_dir=PROJECT_WITH_DUPLICATES,
            config=config,
            packages={},
        )
    l1, l2, l3, l4, l5 = map(str.strip, str(err.value).splitlines())
    assert l1 == "Ambiguous module selected in config.dev.yaml:"
    assert l2 == "module1 exists in:"
    assert l3 == "modules/examples/module1"
    assert l4 == "modules/models/module1"
    assert l5.startswith("You can use the path syntax to disambiguate between modules with the same name")


def test_pull_dataset(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    organization_dir_mutable: Path,
) -> None:
    # Loading a selected dataset to be pulled
    dataset_yaml = organization_dir_mutable / MODULES / "cdf_common" / "data_sets" / "demo.DataSet.yaml"
    dataset = DataSet.load(dataset_yaml.read_text().replace("{{ dataset }}", "ingestion"))
    dataset.description = "New description"
    toolkit_client_approval.append(DataSet, dataset)

    cmd = PullCommand(silent=True)
    cmd.pull_module(
        module_name_or_path=dataset_yaml,
        organization_dir=organization_dir_mutable,
        env="dev",
        dry_run=False,
        verbose=False,
        env_vars=env_vars_with_client,
    )

    reloaded = DataSet.load(dataset_yaml.read_text().replace("{{ dataset }}", "ingestion"))
    assert reloaded.description == "New description"


def test_pull_transformation_sql(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    organization_dir_mutable: Path,
) -> None:
    # Loading a selected transformation to be pulled
    transformation_yaml = (
        organization_dir_mutable
        / "modules"
        / "sourcesystem"
        / "cdf_pi"
        / "transformations"
        / "population"
        / "timeseries.Transformation.yaml"
    )
    source_yaml = transformation_yaml.read_text()
    transformation = _load_cdf_pi_transformation(transformation_yaml, env_vars_with_client.get_client())
    new_query = """select
  someValue as externalId,
  name as name,
  'string' as type,

from `ingestion`.`timeseries_metadata`"""
    transformation.query = new_query

    toolkit_client_approval.append(Transformation, transformation)
    cmd = PullCommand(silent=True)
    cmd.pull_module(
        module_name_or_path=transformation_yaml,
        organization_dir=organization_dir_mutable,
        env="dev",
        dry_run=False,
        verbose=False,
        env_vars=env_vars_with_client,
    )
    sql_file = transformation_yaml.with_suffix(".sql")
    assert sql_file.exists()
    assert sql_file.read_text() == new_query.replace("ingestion", "{{ rawSourceDatabase }}"), "SQL file was not updated"

    target_yaml = transformation_yaml.read_text()
    # Cleanup file endings.
    while target_yaml.endswith("\n"):
        target_yaml = target_yaml[:-1]
    while source_yaml.endswith("\n"):
        source_yaml = source_yaml[:-1]
    assert target_yaml == source_yaml, "Transformation file should not be updated"


def _load_cdf_pi_transformation(transformation_yaml: Path, client: ToolkitClient) -> Transformation:
    variables = [
        ("dataset", "ingestion"),
        ("schemaSpace", "sp_enterprise_process_industry"),
        ("instanceSpace", "springfield_instances"),
        ("organization", "YourOrg"),
        ("timeseriesTransformationExternalId", "pi_timeseries_springfield_aveva_pi"),
        ("sourceName", "Springfield AVEVA PI"),
    ]
    raw_transformation = transformation_yaml.read_text()
    for key, value in variables:
        raw_transformation = raw_transformation.replace(f"{{{{ {key} }}}}", value)
    data = yaml.safe_load(raw_transformation)
    data["dataSetId"] = client.lookup.data_sets.id(data.pop("dataSetExternalId"))
    transformation = Transformation._load(data)

    return transformation


def test_pull_workflow_trigger_with_environment_variables(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    organization_dir_mutable: Path,
) -> None:
    # Loading a selected workflow trigger to be pulled
    yaml_filepath = (
        organization_dir_mutable / "modules" / "cdf_ingestion" / "workflows" / "trigger.WorkflowTrigger.yaml"
    )
    source_yaml = yaml_filepath.read_text()
    vars_replaced = source_yaml
    for key, value in [
        ("{{ workflow }}", "ingestion"),
        # These two secrets are replaced by environment variables
        # that are then replaced with the actual values.
        ("{{ ingestionClientId }}", "this-is-the-ingestion-client-id"),
        ("{{ ingestionClientSecret }}", "this-is-the-ingestion-client-secret"),
    ]:
        vars_replaced = vars_replaced.replace(key, value)
    trigger_dict = yaml.safe_load(vars_replaced)
    trigger_dict["triggerRule"]["cronExpression"] = "* 4 * * *"
    trigger = WorkflowTrigger._load(trigger_dict)
    toolkit_client_approval.append(WorkflowTrigger, trigger)

    cmd = PullCommand(silent=True)
    cmd.pull_module(
        module_name_or_path=yaml_filepath,
        organization_dir=organization_dir_mutable,
        env="dev",
        dry_run=False,
        verbose=False,
        env_vars=env_vars_with_client,
    )
    reloaded = yaml_filepath.read_text()
    assert "cronExpression: '* 4 * * *'" in reloaded, "Workflow trigger was not updated"
    assert "clientId: {{ ingestionClientId }}" in reloaded, "Environment variables were not replaced"
    assert "clientSecret: {{ ingestionClientSecret }}" in reloaded, "Environment variables were not replaced"


def test_pull_group(
    build_tmp_path: Path,
    env_vars_with_client: EnvironmentVariables,
    toolkit_client_approval: ApprovalToolkitClient,
    tmp_path: Path,
) -> None:
    org_dir = tmp_path / "my-org"
    local_file = """name: my_group"""
    local_path = org_dir / "modules" / "my-module" / "auth" / "my_group.Group.yaml"
    local_path.parent.mkdir(parents=True)
    local_path.write_text(local_file)
    cdf_group = Group(
        name="my_group",
        source_id="123-456",
        capabilities=[
            AssetsAcl(scope=AssetsAcl.Scope.All(), actions=[AssetsAcl.Action.Read]),
            TimeSeriesAcl(scope=TimeSeriesAcl.Scope.All(), actions=[TimeSeriesAcl.Action.Read]),
            EventsAcl(scope=EventsAcl.Scope.All(), actions=[EventsAcl.Action.Read]),
        ],
        id=123,
    )
    toolkit_client_approval.append(Group, cdf_group)

    cmd = PullCommand(skip_tracking=True, silent=True)
    cmd.pull_module(
        module_name_or_path="my-module",
        organization_dir=org_dir,
        env="dev",
        dry_run=False,
        verbose=False,
        env_vars=env_vars_with_client,
    )

    reloaded = GroupWrite.load(local_path.read_text())

    assert reloaded.dump() == cdf_group.as_write().dump()


def test_dump_datamodel(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
) -> None:
    # Create a datamodel and append it to the approval client
    space = dm.Space("my_space", is_global=False, last_updated_time=0, created_time=0)
    container = dm.Container(
        space="my_space",
        external_id="my_container",
        name=None,
        description=None,
        properties={"prop1": dm.ContainerProperty(type=dm.Text()), "prop2": dm.ContainerProperty(type=dm.Float64())},
        is_global=False,
        last_updated_time=0,
        created_time=0,
        used_for="node",
        constraints=None,
        indexes=None,
    )
    parent_view = dm.View(
        space="my_space",
        external_id="parent_view",
        version="1",
        properties={
            "prop2": dm.MappedProperty(
                container=container.as_id(),
                container_property_identifier="prop2",
                type=dm.Float64(),
                nullable=True,
                auto_increment=False,
                immutable=False,
            )
        },
        last_updated_time=0,
        created_time=0,
        description=None,
        name=None,
        filter=None,
        implements=None,
        writable=True,
        used_for="node",
        is_global=False,
    )

    view = dm.View(
        space="my_space",
        external_id="my_view",
        version="1",
        properties={
            "prop1": dm.MappedProperty(
                container=container.as_id(),
                container_property_identifier="prop1",
                type=dm.Text(),
                nullable=True,
                auto_increment=False,
                immutable=False,
            ),
        },
        last_updated_time=0,
        created_time=0,
        description=None,
        name=None,
        filter=None,
        implements=[parent_view.as_id()],
        writable=True,
        used_for="node",
        is_global=False,
    )
    data_model = dm.DataModel(
        space="my_space",
        external_id="my_data_model",
        version="1",
        views=[view.as_id(), parent_view.as_id()],
        created_time=0,
        last_updated_time=0,
        description=None,
        name=None,
        is_global=False,
    )
    toolkit_client_approval.append(dm.Space, space)
    toolkit_client_approval.append(dm.Container, container)
    toolkit_client_approval.append(dm.DataModel, data_model)
    toolkit_client_approval.append(dm.View, [parent_view, view])
    cmd = DumpResourceCommand(silent=True)
    cmd.dump_to_yamls(
        DataModelFinder(env_vars_with_client.get_client(), dm.DataModelId.load(("my_space", "my_data_model", "1"))),
        clean=True,
        output_dir=build_tmp_path,
        verbose=False,
    )

    assert len(list(build_tmp_path.glob("**/*.DataModel.yaml"))) == 1
    assert len(list(build_tmp_path.glob("**/*.Container.yaml"))) == 1
    assert len(list(build_tmp_path.glob("**/*.Space.yaml"))) == 1
    view_files = list(build_tmp_path.glob("**/*.View.yaml"))
    assert len(view_files) == 2
    loaded_views = [dm.ViewApply.load(f.read_text()) for f in view_files]
    child_loaded = next(v for v in loaded_views if v.external_id == "my_view")
    assert child_loaded.implements[0] == parent_view.as_id()
    # The parent property should have been removed from the child view.
    assert len(child_loaded.properties) == 1


def test_dump_datamodel_skip_global(
    tmp_path: Path, toolkit_client_approval: ApprovalToolkitClient, env_vars_with_client: EnvironmentVariables
) -> None:
    output_dir = tmp_path / "tmp_dump"
    default_space_args = dict(is_global=False, last_updated_time=0, created_time=0)
    default_view_args = dict(
        last_updated_time=1,
        created_time=1,
        name=None,
        description=None,
        implements=None,
        writable=True,
        used_for="node",
        is_global=False,
        filter=None,
    )
    default_prop_args = dict(
        nullable=True,
        immutable=False,
        auto_increment=False,
        default_value=None,
        name=None,
        description=None,
    )
    default_container_args = dict(
        name=None,
        description=None,
        is_global=False,
        last_updated_time=0,
        created_time=0,
        constraints=None,
        indexes=None,
        used_for="node",
    )
    local_space = dm.Space("my_space", **default_space_args)
    global_space = dm.Space("cdf_cdm", **{**default_space_args, "is_global": True})
    toolkit_client_approval.append(dm.Space, [local_space, global_space])
    local_container = dm.Container(
        space=local_space.space, external_id="MyAsset", properties={}, **default_container_args
    )
    global_container = dm.Container(
        space=global_space.space,
        external_id="CogniteAsset",
        properties={},
        **{**default_container_args, "is_global": True},
    )
    toolkit_client_approval.append(dm.Container, [local_container, global_container])
    local_view = dm.View(
        space=local_space.space,
        external_id="my_view",
        version="1",
        properties={
            "prop1": dm.MappedProperty(
                container=local_container.as_id(),
                container_property_identifier="prop1",
                type=dm.Text(),
                **default_prop_args,
            ),
        },
        **{**default_view_args, "is_global": False},
    )
    global_view = dm.View(
        space=global_space.space,
        external_id="global_view",
        version="1",
        properties={
            "prop1": dm.MappedProperty(
                container=global_container.as_id(),
                container_property_identifier="prop1",
                type=dm.Text(),
                **default_prop_args,
            ),
        },
        **{**default_view_args, "is_global": True},
    )
    toolkit_client_approval.append(dm.View, [local_view, global_view])
    data_model = dm.DataModel(
        space=local_space.space,
        external_id="my_data_model",
        version="1",
        views=[local_view.as_id(), global_view.as_id()],
        is_global=False,
        last_updated_time=0,
        created_time=0,
        name=None,
        description=None,
    )
    toolkit_client_approval.append(dm.DataModel, data_model)

    cmd = DumpResourceCommand(silent=True, skip_tracking=True)
    cmd.dump_to_yamls(
        finder=DataModelFinder(
            env_vars_with_client.get_client(),
            dm.DataModelId.load(("my_space", "my_data_model", "1")),
            include_global=False,
        ),
        clean=True,
        output_dir=output_dir,
        verbose=False,
    )

    assert len(list(output_dir.glob("**/*.DataModel.yaml"))) == 1
    assert len(list(output_dir.glob("**/*.Container.yaml"))) == 1
    assert len(list(output_dir.glob("**/*.Space.yaml"))) == 1
    assert len(list(output_dir.glob("**/*.View.yaml"))) == 1


def test_build_custom_project(
    build_tmp_path: Path,
) -> None:
    expected_resources = {
        "timeseries",
        "data_models",
        "data_sets",
        "raw",
        "extraction_pipelines",
        "transformations",
        "robotics",
    }
    BuildCommand(silent=True).execute(
        organization_dir=PROJECT_NO_COGNITE_MODULES,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=None,
        on_error="raise",
        verbose=False,
    )

    actual_resources = {path.name for path in build_tmp_path.iterdir() if path.is_dir()}

    missing_resources = expected_resources - actual_resources
    assert not missing_resources, f"Missing resources: {missing_resources}"

    extra_resources = actual_resources - expected_resources
    assert not extra_resources, f"Extra resources: {extra_resources}"


def test_build_project_selecting_parent_path(
    build_tmp_path: Path,
) -> None:
    expected_resources = {"auth", "data_models", "files", "transformations", "data_sets"}
    BuildCommand(silent=True).execute(
        organization_dir=PROJECT_FOR_TEST,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=None,
        on_error="raise",
        verbose=False,
    )

    actual_resources = {path.name for path in build_tmp_path.iterdir() if path.is_dir()}

    missing_resources = expected_resources - actual_resources
    assert not missing_resources, f"Missing resources: {missing_resources}"

    extra_resources = actual_resources - expected_resources
    assert not extra_resources, f"Extra resources: {extra_resources}"


def test_deploy_group_with_unknown_acl(
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
) -> None:
    DeployCommand(silent=True).execute(
        env_vars=env_vars_with_client,
        build_dir=BUILD_GROUP_WITH_UNKNOWN_ACL,
        build_env_name="dev",
        drop=False,
        dry_run=False,
        include=None,
        verbose=False,
        drop_data=False,
        force_update=False,
    )

    groups = toolkit_client_approval.created_resources["Group"]
    assert len(groups) == 1
    group = cast(GroupWrite, groups[0])
    assert group.name == "my_group_with_unknown_acl"
    assert len(group.capabilities) == 1
    assert group.capabilities[0].dump() == {
        "someUnknownAcl": {
            "actions": ["UTTERLY_UNKNOWN"],
            "scope": {"unknownScope": {"with": ["some", {"strange": "structure"}]}},
        }
    }


def test_build_project_with_only_top_level_variables(
    build_tmp_path: Path,
) -> None:
    BuildCommand(silent=True).execute(
        organization_dir=PROJECT_NO_COGNITE_MODULES,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="top_level_variables",
        no_clean=False,
        client=None,
        on_error="raise",
        verbose=False,
    )

    assert build_tmp_path.exists()


def test_dump_workflow(
    tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
) -> None:
    # Simulate Workflow, WorkflowVersion, and WorkflowTrigger in CDF.
    toolkit_client_approval.append(
        Workflow,
        Workflow("myWorkflow", 0, 1),
    )
    toolkit_client_approval.append(
        WorkflowVersion,
        WorkflowVersion(
            "myWorkflow",
            "v1",
            WorkflowDefinition(
                # We are testing that the dump fetches the workflow version, not the actual content of the workflow.
                hash_="some-hash",
                tasks=[],
            ),
            0,
            1,
        ),
    )
    toolkit_client_approval.append(
        WorkflowTrigger, WorkflowTrigger("myTrigger", WorkflowScheduledTriggerRule("* * * * * "), "myWorkflow", "v1")
    )

    output_dir = tmp_path / "tmp_dump"
    cmd = DumpResourceCommand(silent=True)
    cmd.dump_to_yamls(
        WorkflowFinder(env_vars_with_client.get_client(), WorkflowVersionId("myWorkflow", "v1")),
        output_dir=output_dir,
        clean=True,
        verbose=False,
    )

    assert len(list(output_dir.glob("**/*.Workflow.yaml"))) == 1
    assert len(list(output_dir.glob("**/*.WorkflowTrigger.yaml"))) == 1
    assert len(list(output_dir.glob("**/*.WorkflowVersion.yaml"))) == 1


def test_build_deploy_location_filter_with_same_filename_in_different_modules(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
) -> None:
    BuildCommand(silent=True).execute(
        False,
        NAUGHTY_PROJECT,
        build_tmp_path,
        ["modules/multi_locations"],
        None,
        False,
        env_vars_with_client.get_client(),
        "raise",
    )

    DeployCommand(silent=True).execute(
        env_vars_with_client,
        build_tmp_path,
        None,
        dry_run=False,
        drop=False,
        drop_data=False,
        force_update=False,
        include=None,
        verbose=False,
    )

    locations = toolkit_client_approval.created_resources_of_type(LocationFilter)

    assert len(locations) == 2


@pytest.mark.skipif(sys.platform != "win32", reason="The encoding issue is only present on Windows")
@pytest.mark.parametrize("encoding", ["utf-8", "cp1252"])
def test_build_deploy_keep_special_characters(
    encoding: str,
    tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
    monkeypatch,
) -> None:
    build_dir = tmp_path / "build"
    build_dir.mkdir(parents=True, exist_ok=True)
    expected_query = "SELECT * FROM my_éñcüd€d£d_table WHERE column = 'value'"

    my_cdf_toml = cdf_toml.CDFToml.load(use_singleton=False)
    my_cdf_toml.cdf.file_encoding = encoding
    monkeypatch.setattr(cdf_toml, "_CDF_TOML", my_cdf_toml)
    BuildCommand(silent=True).execute(
        False, NAUGHTY_PROJECT, build_dir, ["encoding_issue"], None, False, env_vars_with_client.get_client(), "raise"
    )

    DeployCommand(silent=True).execute(
        env_vars_with_client,
        build_dir,
        None,
        dry_run=False,
        drop=False,
        drop_data=False,
        force_update=False,
        include=None,
        verbose=False,
    )

    transformations = toolkit_client_approval.created_resources_of_type(Transformation)

    assert len(transformations) == 2
    transformation = next(t for t in transformations if t.external_id.endswith(encoding))
    assert transformation.query == expected_query


def test_build_project_with_only_identifiers(
    build_tmp_path: Path,
    toolkit_client_approval: ApprovalToolkitClient,
    env_vars_with_client: EnvironmentVariables,
) -> None:
    """In the cdf modules pull command, we have to be able to build a project that only has identifiers
    without raising any errors.
    """
    built_modules = BuildCommand(silent=True, skip_tracking=True).execute(
        verbose=False,
        organization_dir=COMPLETE_ORG_ONLY_IDENTIFIER,
        build_dir=build_tmp_path,
        selected=None,
        build_env_name="dev",
        no_clean=False,
        client=env_vars_with_client.get_client(),
        on_error="raise",
    )

    # Loading the local resources as it is done in the PullCommand
    for loader_cls in RESOURCE_LOADER_LIST:
        loader = loader_cls.create_loader(env_vars_with_client.get_client())
        built_resources = built_modules.get_resources(
            None,
            loader.folder_name,
            loader.kind,
        )
        _ = PullCommand._get_local_resource_dict_by_id(built_resources, loader, {})



from pydantic import Field

from .base import ToolkitResource


class TransformationScheduleYAML(ToolkitResource):
    external_id: str = Field(description="External ID of the scheduled transformation.")
    interval: str = Field(description="Cron expression describes when the job should run.")
    is_paused: bool | None = Field(default=None, description="If true, the transformation is not scheduled.")



from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client import data_modeling as dm
from cognite.client.data_classes import Transformation, TransformationList, TransformationScheduleList
from questionary import Choice

from cognite_toolkit._cdf_tk.client.data_classes.location_filters import LocationFilter, LocationFilterList
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands.dump_resource import (
    DataModelFinder,
    DumpResourceCommand,
    LocationFilterFinder,
    TransformationFinder,
)
from cognite_toolkit._cdf_tk.loaders import LocationFilterLoader, TransformationLoader
from cognite_toolkit._cdf_tk.utils import read_yaml_file
from tests.test_unit.approval_client import ApprovalToolkitClient
from tests.test_unit.utils import MockQuestionary


@pytest.fixture()
def three_transformations() -> TransformationList:
    return TransformationList(
        [
            Transformation(
                1,
                "transformationA",
                name="My First Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
            Transformation(
                2,
                "transformationB",
                name="My Second Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
            Transformation(
                3,
                "transformationC",
                name="My Third Transformation",
                ignore_null_fields=True,
                created_time=1,
                updated_time=1,
            ),
        ]
    )


class TestTransformationFinder:
    def test_select_transformations(self, three_transformations: TransformationList, monkeypatch: MonkeyPatch) -> None:
        def select_transformations(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_transformations)
            return [choices[1].value, choices[2].value]

        answers = [select_transformations]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(TransformationFinder.__module__, monkeypatch, answers),
        ):
            client.transformations.list.return_value = three_transformations
            finder = TransformationFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("transformationB", "transformationC")


class TestDumpTransformations:
    def test_dump_transformations(self, three_transformations: TransformationList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.transformations.retrieve_multiple.return_value = three_transformations[1:]
            client.transformations.schedules.retrieve.return_value = TransformationScheduleList([])

            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                TransformationFinder(client, ("transformationB", "transformationC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = TransformationLoader(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = [read_yaml_file(filepath) for filepath in filepaths]
        assert items == [loader.dump_resource(t) for t in three_transformations[1:]]


class TestDataModelFinder:
    def test_select_data_model(self, toolkit_client_approval: ApprovalToolkitClient, monkeypatch: MonkeyPatch) -> None:
        default_args = dict(
            is_global=False,
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            views=[],
        )
        models = [
            dm.DataModel("my_space", "first_model", "v1", **default_args),
            dm.DataModel("my_space2", "second_model", "v1", **default_args),
        ]
        toolkit_client_approval.append(dm.DataModel, models)
        selected = models[1].as_id()
        finder = DataModelFinder(toolkit_client_approval.mock_client, None)
        answers = ["my_space2", selected, False]
        with MockQuestionary(DataModelFinder.__module__, monkeypatch, answers):
            result = finder._interactive_select()

        assert result == selected
        assert finder.data_model.as_id() == selected


@pytest.fixture()
def three_location_filters() -> LocationFilterList:
    return LocationFilterList(
        [
            LocationFilter(1, external_id="filterA", name="Filter A", created_time=1, updated_time=1),
            LocationFilter(2, external_id="filterB", name="Filter B", created_time=1, updated_time=1),
            LocationFilter(3, external_id="filterC", name="Filter C", created_time=1, updated_time=1),
        ]
    )


class TestLocationFilterFinder:
    def test_select_location_filter(self, three_location_filters: LocationFilterList, monkeypatch: MonkeyPatch) -> None:
        def select_filters(choices: list[Choice]) -> list[str]:
            assert len(choices) == len(three_location_filters)
            return [choices[1].value, choices[2].value]

        answers = [select_filters]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(LocationFilterFinder.__module__, monkeypatch, answers),
        ):
            client.location_filters.list.return_value = three_location_filters
            finder = LocationFilterFinder(client, None)
            selected = finder._interactive_select()

        assert selected == ("filterB", "filterC")


class TestDumpLocationFilter:
    def test_dump_location_filter(self, three_location_filters: LocationFilterList, tmp_path: Path) -> None:
        with monkeypatch_toolkit_client() as client:
            client.location_filters.list.return_value = three_location_filters
            cmd = DumpResourceCommand(silent=True)
            cmd.dump_to_yamls(
                LocationFilterFinder(client, ("filterB", "filterC")),
                output_dir=tmp_path,
                clean=False,
                verbose=False,
            )
            loader = LocationFilterLoader(client, None, None)

        filepaths = list(loader.find_files(tmp_path))
        assert len(filepaths) == 2
        items = [item for filepath in filepaths for item in loader.load_resource_file(filepath)]
        assert items == three_location_filters[1:].as_write().dump()



import sys
from collections.abc import Hashable, Iterator
from dataclasses import dataclass
from typing import Any, Literal, overload
from urllib.parse import urlparse

from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    ClientCredentials,
    OidcCredentials,
)
from cognite.client.data_classes.data_modeling import Edge, Node, ViewId
from cognite.client.data_classes.filters import SpaceFilter
from cognite.client.exceptions import CogniteAPIError
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.constants import ENV_VAR_PATTERN
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
    ToolkitTypeError,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    MediumSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection

from .sql_parser import SQLParser

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


def try_find_error(credentials: OidcCredentials | ClientCredentials | None) -> str | None:
    if credentials is None:
        return None
    missing: list[str] = []
    if client_id := ENV_VAR_PATTERN.match(credentials.client_id):
        missing.append(client_id.group(1))
    if client_secret := ENV_VAR_PATTERN.match(credentials.client_secret):
        missing.append(client_secret.group(1))
    if missing:
        plural = "s are" if len(missing) > 1 else " is"
        return f"The environment variable{plural} not set: {humanize_collection(missing)}."
    if isinstance(credentials, ClientCredentials):
        return None
    try:
        result = urlparse(credentials.token_uri)
        if not all([result.scheme, result.netloc]):
            raise ValueError
    except ValueError:
        return f"The tokenUri {credentials.token_uri!r} is not a valid URI."
    return None


@overload
def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["node"] = "node",
    space: str | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Node]: ...


@overload
def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["edge"],
    space: str | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Edge]: ...


def iterate_instances(
    client: ToolkitClient,
    instance_type: Literal["node", "edge"] = "node",
    space: str | None = None,
    source: ViewId | None = None,
    console: Console | None = None,
) -> Iterator[Node] | Iterator[Edge]:
    """Toolkit specific implementation of the client.data_modeling.instances(...) method to account for 408.

    In addition, we enforce sort based on the argument below (provided by Alex B.).
    """
    body: dict[str, Any] = {"limit": 1_000, "cursor": None, "instanceType": instance_type}
    # Without a sort, the sort is implicitly by the internal id, as cursoring needs a stable sort.
    # By making the sort be on external_id, Postgres should pick the index that's on (project_id, space, external_id)
    # WHERE deleted_at IS NULL. In other words, avoiding soft deleted instances.
    body["sort"] = [
        {
            "property": [instance_type, "externalId"],
            "direction": "ascending",
        }
    ]
    url = f"/api/{client._API_VERSION}/projects/{client.config.project}/models/instances/list"
    if space:
        body["filter"] = SpaceFilter(space=space, instance_type=instance_type).dump()
    if source:
        body["sources"] = [{"source": source.dump(include_type=True, camel_case=True)}]
    while True:
        try:
            response = client.post(url=url, json=body)
        except CogniteAPIError as e:
            if e.code == 408 and body["limit"] > 1:
                MediumSeverityWarning(
                    f"Timeout with limit {body['limit']}, retrying with {body['limit'] // 2}."
                ).print_warning(include_timestamp=True, console=console)
                body["limit"] = body["limit"] // 2
                continue
            raise e
        response_body = response.json()
        if instance_type == "node":
            yield from (Node.load(node) for node in response_body["items"])
        else:
            yield from (Edge.load(edge) for edge in response_body["items"])
        next_cursor = response_body.get("nextCursor")
        if next_cursor is None:
            break
        body["cursor"] = next_cursor


@overload
def read_auth(
    authentication: object,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: Literal[False] = False,
    console: Console | None = None,
) -> ClientCredentials: ...


@overload
def read_auth(
    authentication: object,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: Literal[True],
    console: Console | None = None,
) -> ClientCredentials | OidcCredentials: ...


def read_auth(
    authentication: object,
    client_config: ToolkitClientConfig,
    identifier: Hashable,
    resource_name: str,
    allow_oidc: bool = False,
    console: Console | None = None,
) -> ClientCredentials | OidcCredentials:
    if authentication is None:
        if client_config.is_strict_validation or not isinstance(client_config.credentials, OAuthClientCredentials):
            raise ToolkitRequiredValueError(f"Authentication is missing for {resource_name} {identifier!r}.")
        else:
            HighSeverityWarning(
                f"Authentication is missing for {resource_name} {identifier!r}. Falling back to the Toolkit credentials"
            ).print_warning(console=console)
        return ClientCredentials(client_config.credentials.client_id, client_config.credentials.client_secret)
    elif not isinstance(authentication, dict):
        raise ToolkitTypeError(f"Authentication must be a dictionary for {resource_name} {identifier!r}")
    elif "clientId" not in authentication or "clientSecret" not in authentication:
        raise ToolkitRequiredValueError(
            f"Authentication must contain clientId and clientSecret for {resource_name} {identifier!r}"
        )
    elif allow_oidc and "tokenUri" in authentication and "cdfProjectName" in authentication:
        return OidcCredentials.load(authentication)
    else:
        return ClientCredentials(authentication["clientId"], authentication["clientSecret"])


def get_transformation_sources(query: str) -> list[RawTable | str]:
    """Search the SQL query for source tables."""
    parser = SQLParser(query, operation="Lookup transformation source")
    parser.parse()

    tables: list[RawTable | str] = []
    for table in parser.sources:
        if table.schema == "_cdf":
            tables.append(table.name)
        else:
            tables.append(RawTable(db_name=table.schema, table_name=table.name))
    return tables


def get_transformation_destination_columns(query: str) -> list[str]:
    """Search the SQL query for destination columns."""
    parser = SQLParser(query, operation="Lookup transformation destination columns")
    parser.parse()
    return parser.destination_columns


def metadata_key_counts(
    client: ToolkitClient,
    resource: Literal["assets", "events", "files", "timeseries", "sequences"],
    data_sets: list[int] | None = None,
    hierarchies: list[int] | None = None,
) -> list[tuple[str, int]]:
    """Get the metadata key counts for a given resource.

    Args:
        client: ToolkitClient instance
        resource: The resource to get the metadata key counts for. Can be one of "assets", "events", "files", "timeseries", or "sequences".
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.
        hierarchies: A list of hierarchy IDs to filter by. If None, no filtering is applied.

    Returns:
        A dictionary with the metadata keys as keys and the counts as values.
    """
    where_clause = ""
    if data_sets is not None and hierarchies is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))}) AND rootId IN ({','.join(map(str, hierarchies))})"
    elif data_sets is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))})"
    elif hierarchies is not None:
        where_clause = f"\n         WHERE rootId IN ({','.join(map(str, hierarchies))})"

    query = f"""WITH meta AS (
         SELECT cast_to_strings(metadata) AS metadata_array
         FROM _cdf.{resource}{where_clause}
       ),
       exploded AS (
         SELECT explode(metadata_array) AS json_str
         FROM meta
       ),
       parsed AS (
         SELECT from_json(json_str, 'map<string,string>') AS json_map
         FROM exploded
       ),
       keys_extracted AS (
         SELECT map_keys(json_map) AS keys_array
         FROM parsed
       ),
       all_keys AS (
         SELECT explode(keys_array) AS key
         FROM keys_extracted
       )
       SELECT key, COUNT(key) AS key_count
       FROM all_keys
       GROUP BY key
       ORDER BY key_count DESC;
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    return [(item["key"], item["key_count"]) for item in results.results or []]


def label_count(
    client: ToolkitClient, resource: Literal["assets", "events", "files", "timeseries", "sequences"]
) -> list[dict[str, int | str]]:
    """Get the label counts for a given resource.

    Args:
        client: ToolkitClient instance
        resource: The resource to get the label counts for. Can be one of "assets", "events", "files", "timeseries", or "sequences".

    Returns:
        A dictionary with the labels as keys and the counts as values.
    """
    query = f"""WITH labels as (SELECT explode(labels) AS label
	FROM _cdf.{resource}
  )
SELECT label, COUNT(label) as label_count
FROM labels
GROUP BY label
ORDER BY label_count DESC;
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=1000)
    # We know from the SQL that the result is a list of dictionaries with string keys and int values.
    return results.results or []


@dataclass
class RelationshipCount:
    source_type: str
    target_type: str
    count: int

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(
            source_type=data["sourceType"],
            target_type=data["targetType"],
            count=int(data["relationshipCount"]),
        )


def relationship_aggregate_count(client: ToolkitClient, data_sets: list[int] | None = None) -> list[RelationshipCount]:
    """Get the relationship counts for the asset-centric resources.

    Args:
        client: ToolkitClient instance
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.

    """
    where_clause = ""
    if data_sets is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))})"

    query = f"""SELECT
    sourceType,
    targetType,
    COUNT(externalId) AS relationshipCount
FROM
    _cdf.relationships{where_clause}
GROUP BY
    sourceType,
    targetType
"""
    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    return [
        RelationshipCount(item["sourceType"], item["targetType"], item["relationshipCount"])
        for item in results.results or []
    ]


def label_aggregate_count(client: ToolkitClient, data_sets: list[int] | None = None) -> int:
    """Get the total count of labels in the CDF project.

    Args:
        client: ToolkitClient instance
        data_sets: A list of data set IDs to filter by. If None, no filtering is applied.

    Returns:
        The total count of labels across all resources in the CDF project.
    """
    where_clause = ""
    if data_sets is not None:
        where_clause = f"\n         WHERE dataSetId IN ({','.join(map(str, data_sets))})"

    query = f"""SELECT
    COUNT(externalId) AS labelCount
FROM
    _cdf.labels{where_clause}"""

    results = client.transformations.preview(query, convert_to_string=False, limit=None, source_limit=None)
    if results.results:
        return int(results.results[0]["labelCount"])
    return 0



from collections.abc import Iterable
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteResourceList
from cognite.client.data_classes.data_modeling import EdgeId, NodeId
from cognite.client.data_classes.data_modeling.instances import EdgeApplyResult, InstanceApplyResult, NodeApplyResult


class InstanceApplyResultAdapter(InstanceApplyResult):
    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> InstanceApplyResult:  # type: ignore[override]
        """Load an instance of the class from a resource dictionary."""
        if "instanceType" not in resource:
            raise ValueError("Resource must contain 'instanceType' key.")
        instance_type = resource.get("instanceType")
        if instance_type == "node":
            return NodeApplyResult._load(resource, cognite_client)
        elif instance_type == "edge":
            return EdgeApplyResult._load(resource, cognite_client)
        raise TypeError(f"Resource must be a NodeApplyResult or EdgeApplyResult, not {instance_type}.")


class InstancesApplyResultList(CogniteResourceList[InstanceApplyResult]):
    def __init__(
        self, items: Iterable[InstanceApplyResult] | None = None, cognite_client: CogniteClient | None = None
    ) -> None:
        # Trick to avoid validation in the CogniteResourceList constructor
        super().__init__([], cognite_client=cognite_client)
        self.data.extend(items or [])
        self._build_id_mappings()

    _RESOURCE = InstanceApplyResultAdapter

    def as_ids(self) -> list[NodeId | EdgeId]:
        """Return a list of IDs for the instances in the list."""
        return [
            NodeId(item.space, item.external_id)
            if item.instance_type == "node"
            else EdgeId(item.space, item.external_id)
            for item in self
        ]



from __future__ import annotations

import logging
import tempfile
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadataUpdate, LabelFilter
from cognite.extractorutils.uploader import TimeSeriesUploadQueue
from common.cdf_helpers import create_missing_labels

IMAGE_HEIGHT = 512
IMAGE_WIDTH = 640

logger = logging.getLogger(__name__)


def get_timestamp(file):
    """Get timestamp from file."""
    timestamp = file.metadata.get("timestamp")
    if not timestamp:
        timestamp = file.source_created_time
    if not timestamp:
        # This definitely exists
        timestamp = file.uploaded_time
    return timestamp


def get_asset(client, file):
    """Get asset from metadata field."""
    if file.metadata.get("asset_id", None) not in [None, "0", "None"]:
        try:
            asset_id = int(file.metadata["asset_id"])
        except Exception as e:
            logger.exception(e)
            return None
        return client.assets.retrieve(asset_id)


def handle_failed_upload(client: CogniteClient, id: int, error_message: str, data: dict, metadata: dict | None = None):
    """Log error message and update a file that has failed."""
    print(error_message)

    client.files.update(
        FileMetadataUpdate(id=id)
        .labels.remove(data["input_label"])
        .labels.add([data["output_label"], data["failed_label"]])
        .metadata.add(
            {"error_message": error_message, **metadata} if metadata is not None else {"error_message": error_message}
        )
    )


def handle(data, client):
    print("Start extracting data from the IR raw file.")

    if not {"input_label", "output_label", "success_label", "failed_label", "data_set_external_id"} <= data.keys():
        raise RuntimeError(
            "Data should contain all keys: 'input_label', 'output_label', 'success_label', 'failed_label', 'data_set_external_id'."
        )

    data_set_id = client.data_sets.retrieve(external_id=data["data_set_external_id"]).id

    files = client.files.list(
        labels=LabelFilter(contains_all=[data["input_label"]]),
        limit=-1,
        uploaded=True,
    )

    print(f"Number of files to process with label: {data['input_label']}: {len(files)}.")

    create_missing_labels(client=client, label_ids=[data["output_label"], data["success_label"], data["failed_label"]])

    upload_queue = TimeSeriesUploadQueue(
        client,
        max_upload_interval=1,
        create_missing=True,
        max_queue_size=100,
    )

    for file in files:
        if file.metadata.get("processed") == "false":
            print(f"File with external id: {file.external_id} not processed. Skipping.")
            continue

        ir_image_filename = "ir_image.jpg"
        ir_temperature_filename = "ir_temperatures.csv"
        ir_raw_filename = "ir_raw.raw"

        asset = get_asset(client, file)

        with tempfile.TemporaryDirectory(dir="/tmp") as directory:
            ir_image_path = str(Path.cwd() / directory / ir_image_filename)
            ir_temperature_path = str(Path.cwd() / directory / ir_temperature_filename)
            ir_raw_path = str(Path.cwd() / directory / ir_raw_filename)

            # Download raw file to tmp path
            client.files.download_to_path(path=ir_raw_path, id=file.id)

            # Extract temperatures from the raw file
            temperatures_decikelvin = np.fromfile(file=ir_raw_path, dtype=np.uint16).byteswap()

            # Reshape the raw temperature array
            temperatures_decikelvin = temperatures_decikelvin.reshape((IMAGE_HEIGHT, IMAGE_WIDTH))

            # Convert degrees decikelvin to degrees celsius
            temperatures_celsius = ((temperatures_decikelvin) / 10) - 273.15

            # Save the temperatures to a temp csv file
            np.savetxt(ir_temperature_path, temperatures_celsius, delimiter=",")

            # Save the image to the temp path
            plt.imsave(ir_image_path, temperatures_celsius)

            # Upload image and temperature data to CDF
            try:
                res_image = client.files.upload(
                    path=ir_image_path,
                    external_id=f"ir_image_{file.name}_{file.external_id}_{file.uploaded_time}",
                    name=f"ir_image_{file.external_id}.jpg",
                    data_set_id=data_set_id,
                    mime_type="image/jpeg",
                    asset_ids=[asset.id],
                    metadata={
                        "asset_id": file.metadata.get("asset_id", None),
                        "raw_file_id": file.id,
                        "raw_file_name": file.name,
                    },
                )
                file_update = FileMetadataUpdate(id=file.id).metadata.add({"ir_image_id": res_image.id})
                client.files.update(file_update)

            except Exception as e:
                handle_failed_upload(client=client, id=file.id, error_message=str(e), data=data, metadata=file.metadata)
                continue

            print(f"Uploaded IR image with ID: {res_image.id}.")

            try:
                res_csv = client.files.upload(
                    path=ir_temperature_path,
                    external_id=f"temperatures_{file.name}_{file.external_id}_{file.uploaded_time}",
                    name=f"temperatures_{file.external_id}.csv",
                    data_set_id=data_set_id,
                    mime_type="text/csv",
                    asset_ids=[asset.id],
                    metadata={
                        "asset_id": file.metadata.get("asset_id", None),
                        "raw_file_id": file.id,
                        "raw_file_name": file.name,
                    },
                )
                file_update = FileMetadataUpdate(id=file.id).metadata.add({"ir_temp_csv_id": res_csv.id})
                client.files.update(file_update)

            except Exception as e:
                handle_failed_upload(client=client, id=file.id, error_message=str(e), data=data, metadata=file.metadata)
                continue

            print(f"Uploaded temperature file with ID: {res_csv.id}.")

            # Write the minimum and maximum temperature to the corresponding timeseries
            minimum_temperature = np.amin(temperatures_celsius)
            maximum_temperature = np.amax(temperatures_celsius)

            # if "ts_external_id" in file.metadata:
            if any("ts_external_id" in key for key in file.metadata):
                ts_external_ids = list(value for key, value in file.metadata.items() if "ts_external_id" in key)
                ts_external_ids = sorted(ts_external_ids)

                timestamp = int(get_timestamp(file))

                for ts_eid in ts_external_ids:
                    temp = maximum_temperature if ts_eid.endswith("max") else minimum_temperature
                    eid = ts_eid

                    print(f"Datapoint {temp} at timestamp {timestamp} written to timeseries {ts_eid}.")

                    upload_queue.add_to_upload_queue(
                        datapoints=[(timestamp, temp)],
                        external_id=eid,
                    )

        client.files.update(
            FileMetadataUpdate(id=file.id)
            .labels.remove(data["input_label"])
            .labels.add([data["output_label"], data["success_label"]])
        )

        print(f"IR reading completed successfully for file with external id: {file.external_id}, and id: {file.id}.")

    upload_queue.upload()
    return {}



import json
import math
from dataclasses import dataclass
from pathlib import Path
from typing import Any, cast

import pandas as pd
import questionary
import typer
from cognite.client.data_classes.data_modeling import (
    DataModel,
    MappedProperty,
    NodeApply,
    NodeOrEdgeData,
    PropertyType,
    Space,
    View,
    ViewId,
)
from cognite.client.data_classes.data_modeling.data_types import (
    Boolean,
    CDFExternalIdReference,
    Date,
    DirectRelation,
    Enum,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    ListablePropertyType,
    Text,
    Timestamp,
)
from cognite.client.exceptions import CogniteAPIError
from questionary import Choice
from rich import print
from rich.markup import escape
from rich.progress import Progress

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import (
    CDFAPIError,
    ToolkitFileNotFoundError,
    ToolkitRequiredValueError,
    ToolkitResourceMissingError,
)
from cognite_toolkit._cdf_tk.utils.collection import chunker, humanize_collection
from cognite_toolkit._cdf_tk.utils.file import get_table_columns, read_csv

from ._base import ToolkitCommand


@dataclass
class PopulateConfig:
    view: View
    table: Path
    instance_space: str
    external_id_column: str


class PopulateCommand(ToolkitCommand):
    def view(
        self,
        client: ToolkitClient,
        view_id: list[str] | None = None,
        table: Path | None = None,
        instance_space: str | None = None,
        external_id_column: str | None = None,
        verbose: bool = False,
    ) -> None:
        if view_id is None:
            config = self._get_config_from_user(client)
        else:
            config = self._validate_config(view_id, table, instance_space, external_id_column, client)

        if config.table.suffix == ".csv":
            data = read_csv(config.table)
        else:
            # Parquet - already validated
            data = pd.read_parquet(config.table)

        properties_by_column, property_types_by_column = self._properties_by_column(list(data.columns), config.view)

        with Progress() as progress:
            task = progress.add_task("Populating view", total=len(data))
            for chunk in chunker(data.to_dict(orient="records"), 1_000):
                nodes = [
                    NodeApply(
                        space=config.instance_space,
                        external_id=row[config.external_id_column],
                        sources=[
                            NodeOrEdgeData(
                                source=config.view.as_id(),
                                properties={
                                    properties_by_column[col]: self._serialize_value(
                                        value, property_types_by_column[col], config.instance_space
                                    )
                                    for col, value in row.items()
                                    if col in properties_by_column and col in property_types_by_column
                                },
                            )
                        ],
                    )
                    for row in chunk
                ]
                try:
                    created = client.data_modeling.instances.apply(nodes=nodes, auto_create_direct_relations=True)
                except CogniteAPIError as e:
                    raise CDFAPIError(f"Failed to populate view; {escape(str(e))}")
                else:
                    progress.update(task, advance=len(created.nodes))

    def _get_config_from_user(self, client: ToolkitClient) -> PopulateConfig:
        view = self._get_view_from_user(client)
        table = self._get_table_from_user()
        instance_space = self._get_instance_space_from_user(client)
        external_id_column = self._get_external_id_column_from_user(table)
        return PopulateConfig(
            view=view,
            table=table,
            instance_space=instance_space,
            external_id_column=external_id_column,
        )

    @staticmethod
    def _get_view_from_user(client: ToolkitClient) -> View:
        data_models = client.data_modeling.data_models.list(inline_views=False, limit=-1, all_versions=False)
        data_model_choices = [
            Choice(f"{dm.as_id().as_tuple()}", value=dm)
            for dm in sorted(data_models, key=lambda dm: (dm.space, dm.external_id))
        ]
        selected_data_model: DataModel[ViewId] | None = questionary.select(
            "Select the data model containing the view to populate",
            choices=data_model_choices,
        ).ask()

        if selected_data_model is None:
            print("No data model selected. Exiting.")
            raise typer.Exit(0)

        view_options = [
            Choice(view.external_id, value=view)
            for view in sorted(selected_data_model.views, key=lambda v: v.external_id, reverse=True)
        ]
        selected_view: ViewId | None = questionary.select(
            "Select the view to populate",
            choices=view_options,
        ).ask()
        if selected_view is None:
            print("No view selected. Exiting.")
            raise typer.Exit(0)
        view = client.data_modeling.views.retrieve(selected_view)
        return view[0]

    def _get_table_from_user(self) -> Path:
        selected_table: str | None = questionary.path("Enter the path to the table to populate the view with").ask()
        if selected_table is None:
            print("No table path provided. Exiting.")
            raise typer.Exit(0)
        table_path = Path(selected_table)
        if not table_path.exists():
            print("Table path does not exist.")
            return self._get_table_from_user()
        if table_path.suffix not in (".csv", ".parquet"):
            print("Only CSV and Parquet files are supported. Please provide a valid file.")
            return self._get_table_from_user()
        return table_path

    @staticmethod
    def _get_instance_space_from_user(client: ToolkitClient) -> str:
        spaces = client.data_modeling.spaces.list(limit=-1)
        space_choices = [Choice(space.space, value=space) for space in sorted(spaces, key=lambda s: s.space)]
        selected_space: Space | None = questionary.select(
            "Select the instance space to write the nodes to", choices=space_choices
        ).ask()
        if selected_space is None:
            print("No instance space selected. Exiting.")
            raise typer.Exit(0)
        return selected_space.space

    @staticmethod
    def _get_external_id_column_from_user(table: Path) -> str:
        columns = get_table_columns(table)
        selected_column: str | None = questionary.select(
            "Select the column in the table that contains the external IDs of the nodes",
            choices=[Choice(col, value=col) for col in columns],
        ).ask()
        if selected_column is None:
            print("No external ID column selected. Exiting.")
            raise typer.Exit(0)
        return selected_column

    @staticmethod
    def _validate_config(
        user_view_id: list[str],
        table: Path | None,
        instance_space: str | None,
        external_id_column: str | None,
        client: ToolkitClient,
    ) -> PopulateConfig:
        if missing := [name for name, value in locals().items() if value is None]:
            raise ToolkitRequiredValueError(f"Missing required values: {humanize_collection(missing)}")
        # Happy Mypy
        instance_space = cast(str, instance_space)
        table = cast(Path, table)
        external_id_column = cast(str, external_id_column)

        if not table.exists():
            raise ToolkitFileNotFoundError(f"Table {table.as_posix()} not found", table)
        columns = {col.casefold() for col in get_table_columns(table)}
        if external_id_column.casefold() not in columns:
            raise ToolkitRequiredValueError(
                f"External ID column {external_id_column!r} not found in table {table.name}", external_id_column
            )

        view_id = ViewId.load(tuple(user_view_id))  # type: ignore[arg-type]
        try:
            views = client.data_modeling.views.retrieve(view_id)
        except CogniteAPIError as e:
            raise CDFAPIError(f"Failed to retrieve view {view_id:!r}; {escape(str(e))}")
        if not views:
            raise ToolkitResourceMissingError(f"View {view_id} not found", repr(view_id))
        view = max(views, key=lambda v: v.created_time)
        try:
            space = client.data_modeling.spaces.retrieve(instance_space)
        except CogniteAPIError as e:
            raise CDFAPIError(f"Failed to retrieve instance space {instance_space!r}; {escape(str(e))}")
        if space is None:
            raise ToolkitResourceMissingError(f"Instance space {instance_space} not found", repr(instance_space))

        return PopulateConfig(
            view=view, table=table, instance_space=instance_space, external_id_column=external_id_column
        )

    @staticmethod
    def _properties_by_column(columns: list[str], view: View) -> tuple[dict[str, str], dict[str, PropertyType]]:
        properties_by_column: dict[str, str] = {}
        property_types_by_column: dict[str, PropertyType] = {}
        container_property_by_id = {
            prop_id.casefold(): (prop_id, prop)
            for prop_id, prop in view.properties.items()
            if isinstance(prop, MappedProperty)
        }
        for col in columns:
            if col.casefold() not in container_property_by_id:
                continue
            prop_id, prop = container_property_by_id[col.casefold()]
            properties_by_column[col] = prop_id
            property_types_by_column[col] = prop.type
        return properties_by_column, property_types_by_column

    @classmethod
    def _serialize_value(cls, value: Any, property_type: PropertyType, instance_space: str) -> Any:
        if isinstance(value, str):
            try:
                return cls._serialize_value(json.loads(value), property_type, instance_space)
            except json.JSONDecodeError:
                ...
        elif isinstance(property_type, ListablePropertyType) and property_type.is_list and isinstance(value, list):
            return [cls._serialize_value(v, property_type, instance_space) for v in value]
        elif isinstance(property_type, ListablePropertyType) and property_type.is_list and not isinstance(value, list):
            return [cls._serialize_value(value, property_type, instance_space)]

        if value is None:
            return None

        match (property_type, value):
            case (Text() | CDFExternalIdReference(), _):
                return str(value)
            case (Boolean(), str()):
                return value.lower() in ("true", "1")
            case (Boolean(), _):
                return bool(value)
            case (Timestamp(), _):
                return pd.Timestamp(value).to_pydatetime().isoformat(timespec="milliseconds")
            case (Date(), _):
                return pd.Timestamp(value).strftime("%Y-%m-%d")
            case (Json(), dict() | list() | str()):
                return value
            case (Float32() | Float64(), _):
                float_value = float(value)
                if math.isinf(float_value) or math.isnan(float_value):
                    return None
                return float_value
            case (Int32() | Int64(), _):
                try:
                    return int(value)
                except ValueError:
                    return None
            case (DirectRelation(), _str):
                return {"space": instance_space, "externalId": value}
            case (DirectRelation(), _):
                return value
            case (Enum(), _):
                return next(
                    (
                        opt
                        for opt in property_type.values.keys()
                        if opt.casefold() == str(value).casefold()
                        and str(value).casefold() != (property_type.unknown_value or "").casefold()
                    ),
                    None,
                )
            case _:
                return value
